import ipaddress

from builtins import str

from .base import ConfigObject, ConfigOption
from .command import Command


IPTABLES_WAIT = "5"


def start_iptables_command(cmd, *args):
    return [cmd, "--wait", IPTABLES_WAIT] + list(args)


class ConfigDefaults(ConfigObject):
    typename = "defaults"

    options = [
        ConfigOption(name="input", default="ACCEPT"),
        ConfigOption(name="output", default="ACCEPT"),
        ConfigOption(name="forward", default="ACCEPT"),
        ConfigOption(name="disable_ipv6", type=bool)
    ]

    def getName(self):
        # There should only be one defaults section.  If we return a constant
        # string for all instances, we can match them across file versions.
        return "SINGLETON"

    def get_iptables(self):
        """
        Get the list of iptables commands to use (iptables / ip6tables).
        """
        if self.disable_ipv6:
            return ["iptables"]
        else:
            return ["iptables", "ip6tables"]

    def apply(self, allConfigs):
        commands = list()

        for iptables in self.get_iptables():
            for path in ["input", "output", "forward"]:
                # Create the delegate_X chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "filter", "--new", "delegate_"+path]
                commands.append((self.PRIO_IPTABLES_TOP, Command(cmd, self)))

                # Jump to delegate_X chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "filter", "--append", path.upper(), "--jump",
                        "delegate_"+path]
                commands.append((self.PRIO_IPTABLES_TOP, Command(cmd, self)))

                # Create the X_rule chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "filter", "--new", path+"_rule"]
                commands.append((self.PRIO_IPTABLES_TOP, Command(cmd, self)))

                # Jump to X_rule chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "filter", "--append", path.upper(), "--jump",
                        path+"_rule"]
                commands.append((self.PRIO_IPTABLES_TOP, Command(cmd, self)))

                # Add a rule at the end with the default policy.
                cmd = start_iptables_command(iptables) + ["--table",
                        "filter", "--append", path.upper(), "--jump",
                        getattr(self, path)]
                commands.append((self.PRIO_IPTABLES_TOP, Command(cmd, self)))

            for path in ["prerouting", "postrouting"]:
                # Create the delegate_X chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "nat", "--new", "delegate_"+path]
                commands.append((self.PRIO_IPTABLES_TOP, Command(cmd, self)))

                # Jump to delegate_X chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "nat", "--append", path.upper(), "--jump",
                        "delegate_"+path]
                commands.append((self.PRIO_IPTABLES_TOP, Command(cmd, self)))

                # Create the X_rule chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "nat", "--new", path+"_rule"]
                commands.append((self.PRIO_IPTABLES_TOP, Command(cmd, self)))

                # Jump to X_rule chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "nat", "--append", path.upper(), "--jump",
                        path+"_rule"]
                commands.append((self.PRIO_IPTABLES_TOP, Command(cmd, self)))

        return commands

    def revert(self, allConfigs):
        commands = list()

        for iptables in self.get_iptables():
            for path in ["prerouting", "postrouting"]:
                # Jump to X_rule chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "nat", "--delete", path.upper(), "--jump",
                        path+"_rule"]
                commands.append((-self.PRIO_IPTABLES_TOP, Command(cmd, self)))

                # Flush the X_rule chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "nat", "--flush", path+"_rule"]
                commands.append((-self.PRIO_IPTABLES_TOP, Command(cmd, self)))

                # Delete the X_rule chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "nat", "--delete-chain", path+"_rule"]
                commands.append((-self.PRIO_IPTABLES_TOP, Command(cmd, self)))

                # Jump to delegate_X chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "nat", "--delete", path.upper(), "--jump",
                        "delegate_"+path]
                commands.append((-self.PRIO_IPTABLES_TOP, Command(cmd, self)))

                # Flush the delegate_X chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "nat", "--flush", "delegate_"+path]
                commands.append((-self.PRIO_IPTABLES_TOP, Command(cmd, self)))

                # Delete the delegate_X chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "nat", "--delete-chain", "delegate_"+path]
                commands.append((-self.PRIO_IPTABLES_TOP, Command(cmd, self)))

            for path in ["input", "output", "forward"]:
                # Jump to X_rule chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "filter", "--delete", path.upper(), "--jump",
                        path+"_rule"]
                commands.append((-self.PRIO_IPTABLES_TOP, Command(cmd, self)))

                # Flush the X_rule chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "filter", "--flush", path+"_rule"]
                commands.append((-self.PRIO_IPTABLES_TOP, Command(cmd, self)))

                # Delete the X_rule chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "filter", "--delete-chain", path+"_rule"]
                commands.append((-self.PRIO_IPTABLES_TOP, Command(cmd, self)))

                # Jump to paradrop_X chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "filter", "--delete", path.upper(), "--jump",
                        "delegate_"+path]
                commands.append((-self.PRIO_IPTABLES_TOP, Command(cmd, self)))

                # Flush the paradrop_X chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "filter", "--flush", "delegate_"+path]
                commands.append((-self.PRIO_IPTABLES_TOP, Command(cmd, self)))

                # Delete the paradrop_X chain.
                cmd = start_iptables_command(iptables) + ["--table",
                        "filter", "--delete-chain", "delegate_"+path]
                commands.append((-self.PRIO_IPTABLES_TOP, Command(cmd, self)))

                # Delete the default rule.
                cmd = start_iptables_command(iptables) + ["--table",
                        "filter", "--delete", path.upper(), "--jump",
                        getattr(self, path)]
                commands.append((-self.PRIO_IPTABLES_TOP, Command(cmd, self)))

        return commands

    def updateApply(self, new, allConfigs):
        commands = list()

        for iptables in new.get_iptables():
            for path in ["input", "output", "forward"]:
                if getattr(self, path) == getattr(new, path):
                    # Skip if no change.
                    continue

                # Add the new default rule.
                cmd = start_iptables_command(iptables) + ["--table",
                        "filter", "--append", path.upper(), "--jump",
                        getattr(new, path)]
                commands.append((new.PRIO_IPTABLES_TOP, Command(cmd, new)))

        return commands

    def updateRevert(self, new, allConfigs):
        commands = list()

        for iptables in self.get_iptables():
            for path in ["input", "output", "forward"]:
                if getattr(self, path) == getattr(new, path):
                    # Skip if no change.
                    continue

                # Delete the old default rule.
                cmd = start_iptables_command(iptables) + ["--table",
                        "filter", "--delete", path.upper(), "--jump",
                        getattr(self, path)]
                commands.append((-self.PRIO_IPTABLES_TOP, Command(cmd, self)))

        return commands


class ConfigZone(ConfigObject):
    typename = "zone"

    options = [
        ConfigOption(name="name", required=True),
        ConfigOption(name="network", type=list),
        ConfigOption(name="masq", type=bool, default=False),
        ConfigOption(name="masq_src", type=list, default=["0.0.0.0/0"]),
        ConfigOption(name="masq_dest", type=list, default=["0.0.0.0/0"]),
        ConfigOption(name="conntrack", type=bool, default=False),
        ConfigOption(name="input", default="RETURN"),
        ConfigOption(name="forward", default="RETURN"),
        ConfigOption(name="output", default="RETURN"),
        ConfigOption(name="family", default="any")
    ]

    def get_iptables(self):
        """
        Get the list of iptables commands to use (iptables / ip6tables).
        """
        if self.family == "ipv4":
            return ["iptables"]
        elif self.family == "ipv6":
            return ["ip6tables"]
        else:
            return ["iptables", "ip6tables"]

    def setup(self):
        self._interfaces = list()

    def __commands_iptables(self, allConfigs, action, prio):
        commands = list()

        for interface in self._interfaces:
            for iptables in self.get_iptables():
                # Jump to zone input chain.
                chain = "zone_{}_input".format(self.name)
                cmd = start_iptables_command(iptables) + [
                        "--table", "filter",
                        action, "delegate_input",
                        "--in-interface", interface.config_ifname,
                        "--jump", chain]
                commands.append((prio, Command(cmd, self)))

                # If conntrack is enabled, allow incoming traffic that is
                # associated with allowed outgoing traffic.
                if self.conntrack:
                    comment = "zone {} conntrack".format(self.name)
                    cmd = start_iptables_command(iptables) + [
                            "--table", "filter",
                            action, "input_rule",
                            "--in-interface", interface.config_ifname,
                            "--match", "state", "--state", "ESTABLISHED,RELATED",
                            "--match", "comment", "--comment", comment,
                            "--jump", "ACCEPT"]
                    commands.append((prio, Command(cmd, self)))

                # Implement default policy for zone.
                comment = "zone {} default".format(self.name)
                cmd = start_iptables_command(iptables) + [
                        "--table", "filter",
                        action, "input_rule",
                        "--in-interface", interface.config_ifname,
                        "--match", "comment", "--comment", comment,
                        "--jump", self.input]
                commands.append((prio, Command(cmd, self)))

                # Jump to zone output chain.
                chain = "zone_{}_output".format(self.name)
                cmd = start_iptables_command(iptables) + [
                        "--table", "filter",
                        action, "delegate_output",
                        "--out-interface", interface.config_ifname,
                        "--jump", chain]
                commands.append((prio, Command(cmd, self)))

                # Implement default policy for zone.
                cmd = start_iptables_command(iptables) + [
                        "--table", "filter",
                        action, "output_rule",
                        "--out-interface", interface.config_ifname,
                        "--match", "comment", "--comment", comment,
                        "--jump", self.output]
                commands.append((prio, Command(cmd, self)))

                # Jump to zone forward chain.
                chain = "zone_{}_forward".format(self.name)
                cmd = start_iptables_command(iptables) + [
                        "--table", "filter",
                        action, "delegate_forward",
                        "--in-interface", interface.config_ifname,
                        "--jump", chain]
                commands.append((prio, Command(cmd, self)))

                # Implement default policy for zone.
                cmd = start_iptables_command(iptables) + [
                        "--table", "filter",
                        action, "forward_rule",
                        "--in-interface", interface.config_ifname,
                        "--match", "comment", "--comment", comment,
                        "--jump", self.forward]
                commands.append((prio, Command(cmd, self)))

                # Jump to zone prerouting chain.
                chain = "zone_{}_prerouting".format(self.name)
                cmd = start_iptables_command(iptables) + [
                        "--table", "nat",
                        action, "delegate_prerouting",
                        "--in-interface", interface.config_ifname,
                        "--jump", chain]
                commands.append((prio, Command(cmd, self)))

                # Jump to zone postrouting chain.
                chain = "zone_{}_postrouting".format(self.name)
                cmd = start_iptables_command(iptables) + [
                        "--table", "nat",
                        action, "delegate_postrouting",
                        "--out-interface", interface.config_ifname,
                        "--jump", chain]
                commands.append((prio, Command(cmd, self)))

            # Masquerade rules are for IPv4 only, so add them outside the
            # iptables loop above.
            if self.masq:
                for src in self.masq_src:
                    for dest in self.masq_dest:
                        comment = "zone {} masq".format(self.name)
                        cmd = start_iptables_command("iptables") + [
                                "--table", "nat",
                               action, "POSTROUTING",
                               "--out-interface", interface.config_ifname,
                               "--source", src,
                               "--destination", dest,
                               "--match", "comment", "--comment", comment,
                               "--jump", "MASQUERADE"]
                        commands.append((prio, Command(cmd, self)))

        return commands

    def apply(self, allConfigs):
        # Initialize the list of network:interface sections.
        self._interfaces = list()
        if self.network is not None:
            for networkName in self.network:
                # Look up the interface - may fail.
                interface = self.lookup(allConfigs, "network", "interface", networkName)
                self._interfaces.append(interface)

        commands = list()

        for iptables in self.get_iptables():
            for path in ["input", "output", "forward"]:
                # Create the zone_NAME_X chain.
                chain = "zone_{}_{}".format(self.name, path)
                cmd = start_iptables_command(iptables) + [
                        "--table", "filter",
                        "--new", chain]
                commands.append((self.PRIO_IPTABLES_ZONE, Command(cmd, self)))

            for path in ["prerouting", "postrouting"]:
                # Create the zone_NAME_X chain.
                chain = "zone_{}_{}".format(self.name, path)
                cmd = start_iptables_command(iptables) + [
                        "--table", "nat",
                        "--new", chain]
                commands.append((self.PRIO_IPTABLES_ZONE, Command(cmd, self)))

        commands.extend(self.__commands_iptables(allConfigs, "--append",
                self.PRIO_IPTABLES_ZONE))

        if self.masq:
            self.manager.forwardingCount += 1
            if self.manager.forwardingCount == 1:
                cmd = ["sysctl", "-w", "net.ipv4.conf.all.forwarding=1"]
                commands.append((self.PRIO_IPTABLES_ZONE, Command(cmd, self)))

                cmd = ["sysctl", "-w", "net.ipv6.conf.all.forwarding=1"]
                commands.append((self.PRIO_IPTABLES_ZONE, Command(cmd, self)))

        return commands

    def revert(self, allConfigs):
        commands = self.__commands_iptables(allConfigs, "--delete",
                -self.PRIO_IPTABLES_ZONE)

        if self.masq:
            self.manager.forwardingCount -= 1
            if self.manager.forwardingCount == 0:
                cmd = ["sysctl", "-w", "net.ipv4.conf.all.forwarding=0"]
                commands.append((-self.PRIO_IPTABLES_ZONE, Command(cmd, self)))

                cmd = ["sysctl", "-w", "net.ipv6.conf.all.forwarding=0"]
                commands.append((-self.PRIO_IPTABLES_ZONE, Command(cmd, self)))

        for iptables in self.get_iptables():
            pairs = [
                ("filter", "input"),
                ("filter", "output"),
                ("filter", "forward"),
                ("nat", "prerouting"),
                ("nat", "postrouting")
            ]

            for table, path in pairs:
                chain = "zone_{}_{}".format(self.name, path)

                # Flush the zone_NAME_X chain, so that we do not get an error
                # with the delete command.
                cmd = start_iptables_command(iptables) + [
                        "--table", table,
                        "--flush", chain]
                commands.append((-self.PRIO_IPTABLES_ZONE, Command(cmd, self)))

                # Delete the zone_NAME_X chain.
                cmd = start_iptables_command(iptables) + [
                        "--table", table,
                        "--delete-chain", chain]
                commands.append((-self.PRIO_IPTABLES_ZONE, Command(cmd, self)))

        # Reset the list of network:interface sections.
        self._interfaces = list()

        return commands


class ConfigForwarding(ConfigObject):
    typename = "forwarding"

    options = [
        ConfigOption(name="src", required=True),
        ConfigOption(name="dest", required=True)
    ]

    def __commands(self, allConfigs, action, prio):
        commands = list()

        chain = "zone_{}_forward".format(self.src)
        for dest_iface in self.dest_interfaces:
            comment = "forwarding {} -> {}".format(self.src, self.dest)
            cmd = start_iptables_command("iptables") + [
                    "--table", "filter",
                    action, chain,
                    "--out-interface", dest_iface.config_ifname,
                    "--match", "comment", "--comment", comment,
                    "--jump", "ACCEPT"]
            commands.append((prio, Command(cmd, self)))

        return commands

    def apply(self, allConfigs):
        # Look up src_zone in order to indicate dependency.  If the source zone
        # changes, we need to reapply this forwarding section as well.
        self.src_zone = self.lookup(allConfigs, "firewall", "zone", self.src)

        self.dest_zone = self.lookup(allConfigs, "firewall", "zone", self.dest)

        # Initialize the list of network:interface sections.
        self.dest_interfaces = list()
        if self.dest_zone.network is not None:
            for networkName in self.dest_zone.network:
                # Look up the interface - may fail.
                interface = self.lookup(allConfigs, "network", "interface", networkName)
                self.dest_interfaces.append(interface)

        return self.__commands(allConfigs, "--append", self.PRIO_IPTABLES_RULE)

    def revert(self, allConfigs):
        return self.__commands(allConfigs, "--delete", -self.PRIO_IPTABLES_RULE)


class ConfigRedirect(ConfigObject):
    typename = "redirect"

    options = [
        ConfigOption(name="src"),
        ConfigOption(name="src_ip"),
        ConfigOption(name="src_dip"),
        ConfigOption(name="src_port"),
        ConfigOption(name="src_dport"),
        ConfigOption(name="proto", required=True),
        ConfigOption(name="dest"),
        ConfigOption(name="dest_ip"),
        ConfigOption(name="dest_port"),
        ConfigOption(name="target", default="DNAT")
    ]

    # Any of these values will be interpreted to mean that --proto and ports
    # should not be specified in the iptables rule.
    ANY_PROTO = set([None, "any", "none"])

    def __commands_dnat(self, allConfigs, action, prio):
        """
        Generate DNAT iptables rules.
        """
        commands = list()

        # Special cases:
        # None->skip protocol and port arguments,
        # tcpudp->[tcp, udp]
        if self.proto in self.ANY_PROTO:
            protocols = [None]
        elif self.proto == "tcpudp":
            protocols = ["tcp", "udp"]
        else:
            protocols = [self.proto]

        for proto in protocols:
            chain = "zone_{}_prerouting".format(self.src)
            cmd = start_iptables_command("iptables") + ["--table", "nat",
                   action, chain]

            if self.src_ip is not None:
                cmd.extend(["--source", self.src_ip])
            if self.src_dip is not None:
                cmd.extend(["--destination", self.src_dip])
            if proto is not None:
                cmd.extend(["--proto", proto])
                if self.src_port is not None:
                    cmd.extend(["--sport", self.src_port])
                if self.src_dport is not None:
                    cmd.extend(["--dport", self.src_dport])

            if self.dest_ip is not None:
                if self.dest_port is not None:
                    cmd.extend(["--jump", "DNAT", "--to-destination",
                                "{}:{}".format(
                                    self.dest_ip, self.dest_port)])
                else:
                    cmd.extend(["--jump", "DNAT", "--to-destination",
                               self.dest_ip])
            elif self.dest_port is not None:
                cmd.extend(["--jump", "REDIRECT", "--to-port",
                           self.dest_port])

            commands.append((prio, Command(cmd, self)))

        return commands

    def __commands_snat(self, allConfigs, action, prio):
        """
        Generate SNAT iptables rules.
        """
        commands = list()
        # TODO: implement SNAT rules
        return commands

    def apply(self, allConfigs):
        if self.target == "DNAT":
            commands = self.__commands_dnat(allConfigs, "--insert",
                    self.PRIO_CONFIG_IFACE)
        elif self.target == "SNAT":
            commands = self.__commands_snat(allConfigs, "--insert",
                    self.PRIO_CONFIG_IFACE)
        else:
            raise Exception("Unsupported target ({}) in config {} {}".format(
                self.target, self.typename, self.name))

        self.manager.forwardingCount += 1
        if self.manager.forwardingCount == 1:
            cmd = ["sysctl", "-w", "net.ipv4.conf.all.forwarding=1"]
            commands.append((self.PRIO_CONFIG_IFACE, Command(cmd, self)))

        return commands

    def revert(self, allConfigs):
        if self.target == "DNAT":
            commands = self.__commands_dnat(allConfigs, "--delete",
                    -self.PRIO_CONFIG_IFACE)
        elif self.target == "SNAT":
            commands = self.__commands_snat(allConfigs, "--delete",
                    -self.PRIO_CONFIG_IFACE)
        else:
            commands = list()

        self.manager.forwardingCount -= 1
        if self.manager.forwardingCount == 0:
            cmd = ["sysctl", "-w", "net.ipv4.conf.all.forwarding=0"]
            commands.append((-self.PRIO_CONFIG_IFACE, Command(cmd, self)))

        return commands


class ConfigRule(ConfigObject):
    typename = "rule"

    options = [
        ConfigOption(name="name", required=False),
        ConfigOption(name="src"),
        ConfigOption(name="src_ip"),
        ConfigOption(name="src_port"),
        ConfigOption(name="src_mac"),
        ConfigOption(name="proto"),
        ConfigOption(name="dest"),
        ConfigOption(name="dest_ip"),
        ConfigOption(name="dest_port"),
        ConfigOption(name="target", required=True),
        ConfigOption(name="family", default="any"),
        ConfigOption(name="extra")
    ]

    def get_iptables(self):
        """
        Get the list of iptables commands to use (iptables / ip6tables).
        """
        # Check if the rule explicitly specifies the family.
        if self.family == "ipv4":
            return ["iptables"]
        elif self.family == "ipv6":
            return ["ip6tables"]

        # If the rule specifies the source address, use that as a hint.
        if self.src_ip is not None:
            addr = ipaddress.ip_network(str(self.src_ip), strict=False)
            if addr.version == 4:
                return ["iptables"]
            elif addr.version == 6:
                return ["ip6tables"]

        # If the rule specifies the destination address, use that as a hint.
        if self.dest_ip is not None:
            addr = ipaddress.ip_network(str(self.dest_ip), strict=False)
            if addr.version == 4:
                return ["iptables"]
            elif addr.version == 6:
                return ["ip6tables"]

        # Default is to generate rules for both IPv4 and IPv6.
        return ["iptables", "ip6tables"]

    def __commands(self, allConfigs, action, prio):
        commands = list()

        # Choose INPUT, FORWARD, or OUTPUT based on whether src and dest are
        # specified.
        if self.src and self.dest:
            chain = "zone_{}_forward".format(self.dest)
        elif self.src:
            chain = "zone_{}_input".format(self.src)
        elif self.dest:
            chain = "zone_{}_output".format(self.dest)
        else:
            # TODO: Is this right?
            chain = "input_rule"

        for iptables in self.get_iptables():
            # Put together the common options for rules.
            cmd = start_iptables_command(iptables) + [
                    "--wait", IPTABLES_WAIT,
                    "--table", "filter",
                    action, chain]

            # Source and destination IP address.
            if self.src_ip:
                cmd.extend(["--source", self.src_ip])
            if self.dest_ip:
                cmd.extend(["--destination", self.dest_ip])

            # Source and destination ports.
            if self.proto:
                cmd.extend(["--protocol", self.proto])
                if self.src_port:
                    cmd.extend(["--sport", str(self.src_port)])
                if self.dest_port:
                    cmd.extend(["--dport", str(self.dest_port)])

            # Source MAC address.
            if self.src_mac:
                cmd.extend(["--match", "mac", "--mac-source", self.src_mac])

            # Extra options such as "--match state --state ESTABLISHED,RELATED"
            if self.extra:
                cmd.append(self.extra)

            cmd.extend(["--jump", self.target])

            if self.name:
                cmd.extend(["--match", "comment", "--comment", self.name])

            commands.append((prio, Command(cmd, self)))

        return commands

    def apply(self, allConfigs):
        return self.__commands(allConfigs, "--append", self.PRIO_IPTABLES_RULE)

    def revert(self, allConfigs):
        return self.__commands(allConfigs, "--delete", -self.PRIO_IPTABLES_RULE)



import ipaddress

from builtins import str
from http.server import BaseHTTPRequestHandler

import six

from twisted.internet.protocol import DatagramProtocol

from paradrop.base import settings
from paradrop.core.system import system_info


SSDP_ADDR = "239.255.255.250"

PARADROP_URN = "urn:schemas-upnp-org:service:Paradrop:1"


# Warning: SSDP can be used for traffic amplification attacks.
# See this for example: https://blog.cloudflare.com/ssdp-100gbps/
# We do not want Paradrop to be used for DoS attacks, so here are
# a couple of ways to guard against malicious behavior.
#
# 1. Configure iptables to block UDP port 1900 on the WAN interface.
#    This is up to the device owner to configure properly.
# 2. Only send responses to requests that came from private addresses,
#    as these should be LAN devices. This approach is implemented here.


class SsdpRequest(BaseHTTPRequestHandler):
    def __init__(self, request_text):
        self.rfile = six.StringIO(request_text)
        self.raw_requestline = self.rfile.readline()
        self.error_code = None
        self.error_message = None
        self.parse_request()

    def send_error(self, code, message):
        self.error_code = code
        self.error_message = message


class SsdpResponder(DatagramProtocol):
    def startProtocol(self):
        self.transport.joinGroup(SSDP_ADDR)

    def datagramReceived(self, datagram, address):
        request = SsdpRequest(datagram)
        addr = ipaddress.ip_address(str(address[0]))

        # Since we really only need to support queries from clients based on
        # pdtools, we can be choosy about what requests we will respond to.
        #
        # 1. Only reply to M-SEARCH requests.
        # 2. The search target (ST) must match our URN or one of the common
        #    wildcard targets.
        # 3. The recipient MUST be a private IP address.
        if not request.error_code and \
                request.command == "M-SEARCH" and \
                request.path == "*" and \
                request.headers['ST'] in ["ssdp:all", "upnp:rootdevice", PARADROP_URN] and \
                request.headers['MAN'] == '"ssdp:discover"' and \
                addr.is_private:
            os_ver = system_info.getOSVersion()
            pd_ver = system_info.getPackageVersion("paradrop")

            message = "\r\n".join([
                "HTTP/1.1 200 OK",
                "CACHE-CONTROL: max-age=60",
                "LOCATION: http://{domain}",
                "SERVER: OS/{os_ver} UPnP/1.1 Paradrop/{pd_ver}",
                "ST: {urn}"
            ]).format(domain=settings.LOCAL_DOMAIN, os_ver=os_ver,
                    pd_ver=pd_ver, urn=PARADROP_URN)

            self.transport.write(message, address)

    def stop(self):
        self.ssdp.leaveGroup(SSDP_ADDR)
        self.ssdp.stopListening()



class Step(object):
    """
    Defines a step in an update execution plan.  Steps can have inputs/outputs
    and dependencies that define their ordering as well as conditions that
    determine whether they should be applied.
    """

    def __init__(self, func, gets=[], sets=[]):
        self.func = func
        self.gets = set(gets)
        self.sets = set(sets)

    def __cmp__(self, other):
        if (self.sets & other.gets):
            return -1
        elif (other.sets & self.gets):
            return 1
        else:
            return 0


class AbortStep(Step):
    """
    A specific execution step that should be applied during abort sequence.
    """

    def __init__(self, func, target):
        super(AbortStep, self).__init__(func)



import attr


@attr.s
class Service(object):
    """
    A service is a long-running process that provides chute functionality.
    """

    command     = attr.ib(default=None)
    dockerfile  = attr.ib(default=None)
    image       = attr.ib(default=None)
    name        = attr.ib(default=None)
    source      = attr.ib(default=None)
    type        = attr.ib(default="normal")
    build       = attr.ib(default=attr.Factory(dict))
    environment = attr.ib(default=attr.Factory(dict))
    interfaces  = attr.ib(default=attr.Factory(dict))
    requests    = attr.ib(default=attr.Factory(dict))

    # Reference to parent chute. This needs to be private (starts with an
    # underscore); otherwise, we have a circular reference problem when trying
    # to serialize the chute.
    _chute      = attr.ib(default=None)

    def create_specification(self):
        """
        Create a new service specification.

        This is a completely clean copy of all information necessary to rebuild
        the Service object. It should contain only primitive types, which can
        easily be serialized as JSON or YAML.
        """
        spec = {
            "type": self.type,
            "source": self.source,
            "image": self.image,
            "command": self.command,
            "build": self.build.copy(),
            "environment": self.environment.copy(),
            "interfaces": self.interfaces.copy(),
            "requests": self.requests.copy()
        }
        return spec

    def get_chute(self):
        """
        Get the chute to which this service belongs.
        """
        return self._chute

    def get_container_name(self):
        """
        Get the name for the service's container.

        This will be a combination of the chute name and the service name.
        """
        if self.name is None:
            # name can be None for old-style single-service chutes where the
            # container name is expected to be the name of the chute.
            return self._chute.name
        else:
            return "{}-{}".format(self._chute.name, self.name)

    def get_image_name(self):
        """
        Get the name of the image to be used.
        """
        # Light chute services have a shorthand image name like "python2" that
        # should not be interpreted as an actual Docker image name.
        if self.image is None or self.type == "light":
            return "{}:{}".format(self.get_container_name(), self._chute.version)
        else:
            return self.image



from autobahn.twisted.websocket import WebSocketServerProtocol
from twisted.internet import interfaces
from zope.interface import implementer


@implementer(interfaces.IConsumer)
@implementer(interfaces.IPushProducer)
class ProducerConsumerWsProtocol(WebSocketServerProtocol):
    """
    Generic producer/consumer websocket protocol.

    This protocol accepts data from any implementer of IPushProducer or
    IPullProducer and forwards it to the websocket connection.  It also reads
    data from the websocket connection and forwards it to any implementer of
    IConsumer.  Either of the consumer or producer sides are optional.
    """
    def __init__(self, consumer=None):
        WebSocketServerProtocol.__init__(self)
        self.connected = False
        self.producer = None
        self.streaming = True

        self.consumer = consumer
        self.paused = False

    def onClose(self, wasClean, code, reason):
        self.connected = False
        if self.producer is not None:
            self.producer.stopProducing()

    def onOpen(self):
        self.connected = True
        if self.producer is not None:
            self.producer.resumeProducing()

    def onMessage(self, payload, isBinary):
        if self.consumer is not None:
            self.consumer.write(payload)

    #
    # IConsumer interface
    #

    def registerProducer(self, producer, streaming):
        self.producer = producer
        self.streaming = streaming

        # We are ready to receive data from the producer when the websocket
        # connection is open.
        if self.connected:
            producer.resumeProducing()
        else:
            producer.pauseProducing()

    def unregisterProducer(self):
        self.producer = None

    def write(self, data):
        self.sendMessage(data)
        if not self.streaming:
            self.producer.resumeProducing()

    #
    # IPushProducer interface
    #

    def pauseProducing(self):
        self.paused = True
        if self.connected:
            self.transport.pauseProducing()

    def resumeProducing(self):
        self.paused = False
        if self.connected:
            self.transport.resumeProducing()

    def stopProducing(self):
        if self.connected:
            self.transport.stopProducing()



import os
from setuptools import setup, find_packages

with open("../../README.md", "r") as source:
    long_description = source.read()

def package_files(parent):
    files = []
    for (dirpath, dirnames, filenames) in os.walk(parent):
        for f in filenames:
            files.append(os.path.join(dirpath, f))
    return files

setup(
    name="paradrop",
    version='0.13.3a4',
    description="Paradrop Edge Computing Agent",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://paradrop.org",

    project_urls = {
        "Documentation": "https://paradrop.readthedocs.io/en/latest/",
        "Homepage": "https://paradrop.org",
        "Source": "https://github.com/ParadropLabs/Paradrop",
    },

    packages=find_packages(),

    install_requires=[
        'attrs==18.2.0',
        'autobahn==18.8.1',
        'bcrypt==3.1.4',
        'colorama==0.3.9',
        'configparser==3.7.1',
        'docker==2.5.1',
        'enum34',
        'functools32==3.2.3.post2; python_version < "3.2"',
        'future==0.16.0',
        'ipaddress>=1.0.16',
        'jsonpatch==1.15',
        'klein==16.12.0',
        'netifaces==0.10.4',
        'psutil==5.6.2',
        'pulsectl==17.12.2',
        'pycurl',
        'PyJWT==1.6.4',
        'PyYAML==5.1',
        'requests==2.21.0',
        'ruamel.ordereddict==0.4.13; python_version < "3.1"',
        'ruamel.yaml==0.15.60',
        'service_identity==17.0.0',
        'six>=1.10.0',
        'smokesignal==0.7.0',
        'Twisted==16.2.0',
        'wget==3.2'
    ],

    python_requires=">=2.7",

    classifiers = [
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development",
        "Topic :: Utilities",
    ],

    maintainer="Lance Hartung",
    maintainer_email="lance@paradrop.io",

    data_files=[
        ("www", package_files("../localweb/www")),
    ],

    entry_points={
        'console_scripts': [
            'paradrop = paradrop:main',
            'pdconfd = paradrop:run_pdconfd'
        ]
    },
)



"""
This module is responsible for configuration haproxy.
"""
import os
import subprocess

from paradrop.base import settings
from paradrop.core.chute.chute_storage import ChuteStorage
from paradrop.core.container.chutecontainer import ChuteContainer


def generateConfigSections():
    sections = []

    sections.append({
        "header": "global",
        "lines": [
            "daemon",
            "maxconn 256",
        ]
    })

    sections.append({
        "header": "defaults",
        "lines": [
            "mode http",
            "timeout connect 5000ms",
            "timeout client 50000ms",
            "timeout server 50000ms"
        ]
    })

    sections.append({
        "header": "backend portal",
        "lines": [
            "server pd_portal 127.0.0.1:8080 maxconn 256"
        ]
    })

    # Custom variables:
    # - req.querymarker: will be set to the literal "?" if the original request
    # contains a query string.  We will use this to construct a redirect with a
    # query string only if needed.
    # - req.subpath: will be set to the remainder of the path, if anything,
    # after removing /chutes/<chutename>, e.g. "/chutes/hello-world/index.html"
    # becomes "/index.html".  This does not include the query string.
    frontend = {
        "header": "frontend http-in",
        "lines": [
            "bind *:80",
            "default_backend portal",
            "http-request set-var(req.querymarker) str(?) if { query -m found }",
            "http-request set-var(req.subpath) path,regsub(^/chutes/[^/]+,)"
        ]
    }
    sections.append(frontend)

    chuteStore = ChuteStorage()
    chutes = chuteStore.getChuteList()
    for chute in chutes:
        port, service = chute.get_web_port_and_service()
        if port is None or service is None:
            continue

        container = ChuteContainer(service.get_container_name())
        if not container.isRunning():
            continue

        # Generate a rule that matches HTTP host header to chute name.
        frontend['lines'].append("acl host_{} hdr(host) -i {}.chute.paradrop.org".format(
            chute.name, chute.name))
        frontend['lines'].append("use_backend {} if host_{}".format(
            chute.name, chute.name))

        # Generate rules that matches the beginning of the URL.
        # We need to be careful and either have an exact match
        # or make sure there is a slash or question mark after the chute name
        # to avoid mix-ups, e.g. "sticky-board" and "sticky-board-new".
        frontend['lines'].append("acl path_{} url /chutes/{}".format(
            chute.name, chute.name))
        frontend['lines'].append("acl path_{} url_beg /chutes/{}/".format(
            chute.name, chute.name))
        frontend['lines'].append("acl path_{} url_beg /chutes/{}?".format(
            chute.name, chute.name))

        # Try to find a host binding for the web port to redirect:
        # http://<host addr>/chutes/<chute>/<path> ->
        # http://<host addr>:<chute port>/<path>
        #
        # We need to do a lookup because the host port might be dynamically
        # assigned by Docker.
        #
        # Use HTTP code 302 for the redirect, which will not be cached by the
        # web browser.  The port portion of the URL can change whenever the
        # chute restarts, so we don't want web browsers to cache it.  Browsers
        # will cache a 301 (Moved Permanently) response.
        portconf = container.getPortConfiguration(port, "tcp")
        if len(portconf) > 0:
            # TODO: Are there other elements in the list?
            binding = portconf[0]
            frontend['lines'].append("http-request replace-value Host (.*):(.*) \\1")
            frontend['lines'].append("http-request redirect location http://%[req.hdr(host)]:{}%[var(req.subpath)]%[var(req.querymarker)]%[query] code 302 if path_{}".format(
                binding['HostPort'], chute.name))

        # Add a server at the chute's IP address.
        sections.append({
            "header": "backend {}".format(chute.name),
            "lines": [
                "server {} {}:{} maxconn 256".format(chute.name,
                    container.getIP(), port)
            ]
        })

    return sections


def writeConfigFile(output):
    sections = generateConfigSections()
    for section in sections:
        output.write(section['header'] + "\n")
        for line in section['lines']:
            output.write("    " + line + "\n")
        output.write("\n")


def reconfigureProxy(update):
    """
    Reconfigure haproxy with forwarding and redirect rules.
    """
    confFile = os.path.join(settings.RUNTIME_HOME_DIR, "haproxy.conf")
    pidFile = os.path.join(settings.TMP_DIR, "haproxy.pid")

    with open(confFile, "w") as output:
        writeConfigFile(output)

    cmd = ["haproxy", "-f", confFile, "-p", pidFile]

    if os.path.exists(pidFile):
        with open(pidFile, "r") as source:
            pid = source.read().strip()
            cmd.extend(["-sf", pid])

    subprocess.call(cmd)



"""
Manage audio input and output.

Endpoints for these functions can be found under /api/v1/audio.
"""

import os
import json

from klein import Klein

import pulsectl

from . import cors


PULSE_CLIENT_NAME = os.environ.get('SNAP_NAME', 'paradrop')


class AudioApi(object):
    routes = Klein()

    @routes.route('/info', methods=['GET'])
    def get_info(self, request):
        """
        Get audio server information.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/audio/info

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             "default_sink_name": "alsa_output.usb-Performance_Designed_Products_PDP_Audio_Device-00.analog-stereo",
             "default_source_name": "alsa_input.usb-Performance_Designed_Products_PDP_Audio_Device-00.analog-mono",
             "host_name": "localhost.localdomain",
             "server_name": "pulseaudio",
             "server_version": "8.0-rebootstrapped",
             "user_name": "root"
           }
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        with pulsectl.Pulse(PULSE_CLIENT_NAME) as pulse:
            info = pulse.server_info()

        result = {
            'default_sink_name': info.default_sink_name,
            'default_source_name': info.default_source_name,
            'host_name': info.host_name,
            'server_name': info.server_name,
            'server_version': info.server_version,
            'user_name': info.user_name
        }
        return json.dumps(result)

    @routes.route('/info', methods=['PUT'])
    def update_info(self, request):
        """
        Update audio server information.

        The only writable attributes are default_sink_name and default_source_name.

        **Example request**:

        .. sourcecode:: http

           PUT /api/v1/audio/info

           {
             "default_sink_name": "alsa_output.usb-Performance_Designed_Products_PDP_Audio_Device-00.analog-stereo",
             "default_source_name": "alsa_input.usb-Performance_Designed_Products_PDP_Audio_Device-00.analog-mono"
           }

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             "default_sink_name": "alsa_output.usb-Performance_Designed_Products_PDP_Audio_Device-00.analog-stereo",
             "default_source_name": "alsa_input.usb-Performance_Designed_Products_PDP_Audio_Device-00.analog-mono"
           }
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        body = json.loads(request.content.read())

        result = {}
        with pulsectl.Pulse('paradrop-daemon') as pulse:
            if 'default_sink_name' in body:
                pulse.sink_default_set(body['default_sink_name'])
                result['default_sink_name'] = body['default_sink_name']
            if 'default_source_name' in body:
                pulse.source_default_set(body['default_source_name'])
                result['default_source_name'] = body['default_source_name']

        return json.dumps(result)

    @routes.route('/modules', methods=['GET'])
    def get_modules(self, request):
        """
        List loaded modules.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/audio/modules

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           [
             { "index": 0, "n_used": 4294967295, "name": "module-device-restore" },
             { "index": 1, "n_used": 4294967295, "name": "module-stream-restore" },
             { "index": 2, "n_used": 4294967295, "name": "module-card-restore" },
             { "index": 3, "n_used": 4294967295, "name": "module-augment-properties" },
             ...
           ]
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        result = []

        with pulsectl.Pulse('paradrop-daemon') as pulse:
            for module in pulse.module_list():
                result.append({
                    'index': module.index,
                    'name': module.name,
                    'n_used': module.n_used
                })

        return json.dumps(result)

    @routes.route('/modules', methods=['POST'])
    def load_module(self, request):
        """
        Load a module.

        **Example request**:

        .. sourcecode:: http

           POST /api/v1/audio/modules

           {
             "name": "module-switch-on-connect"
           }

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           {
             "index": 21,
             "name": "module-switch-on-connect"
           }
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        body = json.loads(request.content.read())

        with pulsectl.Pulse('paradrop-daemon') as pulse:
            index = pulse.module_load(body['name'])

        result = {
            'index': index,
            'name': body['name']
        }
        return json.dumps(result)

    @routes.route('/sinks', methods=['GET'])
    def get_sinks(self, request):
        """
        List sinks.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/audio/sinks

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           [
             {
               "channel_count": 2,
               "channel_list": ["front-left", "front-right"],
               "description": "PDP Audio Device Analog Stereo",
               "index": 0,
               "name": "alsa_output.usb-Performance_Designed_Products_PDP_Audio_Device-00.analog-stereo",
               "volume": [1.5, 1.5]
             }
           ]
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        result = []
        with pulsectl.Pulse('paradrop-daemon') as pulse:
            for sink in pulse.sink_list():
                result.append({
                    'channel_count': sink.channel_count,
                    'channel_list': sink.channel_list,
                    'description': sink.description,
                    'index': sink.index,
                    'name': sink.name,
                    'volume': sink.volume.values
                })

        return json.dumps(result)

    @routes.route('/sinks/<string:name>/volume', methods=['PUT'])
    def set_sink_volume(self, request, name):
        """
        Set sink volume.

        **Example request**:

        .. sourcecode:: http

           PUT /api/v1/audio/sinks/alsa_output.usb-Performance_Designed_Products_PDP_Audio_Device-00.analog-stereo/volume

           [1.0, 1.0]

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           [1.0, 1.0]
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        body = json.loads(request.content.read())

        with pulsectl.Pulse('paradrop-daemon') as pulse:
            found = False
            for sink in pulse.sink_list():
                if sink.name == name:
                    found = True
                    break

            if found:
                volume = pulsectl.PulseVolumeInfo(body)
                pulse.volume_set(sink, volume)
                return json.dumps(body)
            else:
                request.setResponseCode(404)
                return '{}'

    @routes.route('/sources', methods=['GET'])
    def get_sources(self, request):
        """
        List audio sources.

        **Example request**:

        .. sourcecode:: http

           GET /api/v1/audio/sources

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           [
             {
               "channel_count": 2,
               "channel_list": ["front-left", "front-right"],
               "description": "Monitor of PDP Audio Device Analog Stereo",
               "index": 0,
               "name": "alsa_output.usb-Performance_Designed_Products_PDP_Audio_Device-00.analog-stereo.monitor",
               "volume": [1.0, 1.0]
             },
             {
               "channel_count": 1,
               "channel_list": ["mono"],
               "description": "PDP Audio Device Analog Mono",
               "index": 1,
               "name": "alsa_input.usb-Performance_Designed_Products_PDP_Audio_Device-00.analog-mono",
               "volume": [0.4298553466796875]
             }
           ]
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        result = []
        with pulsectl.Pulse('paradrop-daemon') as pulse:
            for source in pulse.source_list():
                result.append({
                    'channel_count': source.channel_count,
                    'channel_list': source.channel_list,
                    'description': source.description,
                    'index': source.index,
                    'name': source.name,
                    'volume': source.volume.values
                })

        return json.dumps(result)

    @routes.route('/sources/<string:name>/volume', methods=['PUT'])
    def set_source_volume(self, request, name):
        """
        Set source volume.

        **Example request**:

        .. sourcecode:: http

           PUT /api/v1/audio/sources/alsa_input.usb-Performance_Designed_Products_PDP_Audio_Device-00.analog-mono/volume

           [1.0]

        **Example response**:

        .. sourcecode:: http

           HTTP/1.1 200 OK
           Content-Type: application/json

           [1.0]
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        body = json.loads(request.content.read())

        with pulsectl.Pulse('paradrop-daemon') as pulse:
            found = False
            for source in pulse.source_list():
                if source.name == name:
                    found = True
                    break

            if found:
                volume = pulsectl.PulseVolumeInfo(body)
                pulse.volume_set(source, volume)
                return json.dumps(body)
            else:
                request.setResponseCode(404)
                return '{}'



import errno
import json
import os
import time

from http.client import HTTPConnection

from paradrop.base import settings
from paradrop.base.output import out
from paradrop.lib.utils import datastruct


def wait_for_zerotier(max_delay=120):
    """
    Wait for ZeroTier to start up and create the authtoken file.
    """
    path = os.path.join(settings.ZEROTIER_LIB_DIR, "authtoken.secret")

    if not os.path.isfile(path):
        out.info("Waiting up to {} seconds for ZeroTier authtoken\n".format(max_delay))

    while not os.path.isfile(path):
        time.sleep(1)


def configure(update):
    hostConfig = update.cache_get('hostConfig')

    # Check if ZeroTier is enabled in the node configuration and also whether
    # its data directory exists. We prefer not to wait for it if it is not even
    # installed.
    enabled = datastruct.getValue(hostConfig, "zerotier.enabled", False)
    if enabled and os.path.isdir(settings.ZEROTIER_LIB_DIR):
        new_networks = set(datastruct.getValue(hostConfig,
                "zerotier.networks", []))
        wait_for_zerotier()
    else:
        new_networks = set()

    old_networks = set()
    for network in get_networks(ignore_error=not enabled):
        old_networks.add(network['id'])

    # Leave old networks and join new networks.
    for net in (old_networks - new_networks):
        manage_network(net, action="leave")
    for net in (new_networks - old_networks):
        manage_network(net, action="join")


def getAddress():
    """
    Return the zerotier address for this device or None if unavailable.
    """
    path = os.path.join(settings.ZEROTIER_LIB_DIR, "identity.public")
    try:
        with open(path, "r") as source:
            return source.read().strip()
    except IOError as error:
        if error.errno == errno.ENOENT:
            return None
        else:
            raise


def get_auth_token():
    """
    Return the zerotier auth token for accessing its API or None if unavailable.
    """
    path = os.path.join(settings.ZEROTIER_LIB_DIR, "authtoken.secret")
    try:
        with open(path, "r") as source:
            return source.read().strip()
    except IOError as error:
        if error.errno == errno.ENOENT:
            return None
        else:
            raise


def get_networks(ignore_error=False):
    """
    Get list of active ZeroTier networks.
    """
    auth_token = get_auth_token()
    if auth_token is None:
        return []

    conn = HTTPConnection("localhost", 9993)
    path = "/network"
    headers = {
        "X-ZT1-Auth": auth_token
    }
    conn.request("GET", path, "", headers)
    res = conn.getresponse()
    data = json.loads(res.read())

    # nwid field is deprecated, so make sure id field exists.
    for network in data:
        if 'id' not in network and 'nwid' in network:
            network['id'] = network['nwid']

    return data


def manage_network(nwid, action="join"):
    """
    Join or leave a ZeroTier network.

    nwid: ZeroTier network ID, e.g. "e5cd7a9e1c8a5e83"
    action: either "join" or "leave"
    """
    if action == "join":
        method = "POST"
    elif action == "leave":
        method = "DELETE"
    else:
        raise Exception("Unsupported action: {}".format(action))

    conn = HTTPConnection("localhost", 9993)
    path = "/network/{}".format(nwid)
    body = "{}"
    headers = {
        "Content-Type": "application/json",
        "X-ZT1-Auth": get_auth_token()
    }
    conn.request(method, path, body, headers)
    res = conn.getresponse()
    data = json.loads(res.read())
    return data



import six

from .base import ConfigObject, ConfigOption
from .command import Command


# Map class names to priority band numbers (0 is maximum priority).
DEFAULT_CLASSES = {
    "BestEffort": 2,
    "Bulk": 3,
    "Interactive": 1,
    "IntBulk": 2,
    "Priority": 0
}


def compute_hfsc_params(classes, capacity):
    result = dict()

    sum_priority = 0
    sum_avgrate = 0
    sum_limitrate = 0

    # During the first pass, calculate the sums of rates and priorities and
    # determine which service curves are present for each class.
    for cid, cl in classes:
        params = {
            "has_rt": False,
            "has_ls": False,
            "has_ul": False,
            "has_bend": False
        }

        # The avgrate option adds real-time (rt) service curve.
        if cl.avgrate is not None:
            sum_avgrate += cl.avgrate
            params['has_rt'] = True

        # The priority option adds link share (ls) service curve.
        if cl.priority is not None:
            sum_priority += cl.priority
            params['has_ls'] = True

        if cl.limitrate is not None:
            sum_limitrate += cl.limitrate
            params['has_ul'] = True

        # Note: if the class has neither rt nor ls curves,
        # it will get no service (packets dropped).

        # The packetsize or packetdelay are not very intuitive names, but we
        # use them to set the duration of the first bend of the service curve.
        if (cl.packetsize is not None or cl.packetdelay is not None):
            params['has_bend'] = True

        result[cid] = params

    for cid, cl in classes:
        params = result[cid]

        if params['has_rt']:
            # In steady state, each class with a real-time curve gets at least
            # its allocated rate. The avgrate option is interpreted as a
            # percentage of the capacity.
            params['rt_m2'] = float(cl.avgrate) * capacity / 100.0

            if params['has_bend']:
                # The classes with real-time curves are allowed to temporarily
                # burst up to the full capacity.
                params['rt_m1'] = float(cl.avgrate) / sum_avgrate * capacity

                params['rt_d'] = cl.packetdelay
                if cl.packetsize is not None:
                    transfer_delay = float(8 * cl.packetsize) / params['rt_m1']
                    params['rt_d'] = max(params['rt_d'], transfer_delay)

        if params['has_ls']:
            # In steady state, each class with a link-share curve gets a
            # proportion of the capacity defined by its priority.
            params['ls_m2'] = float(cl.priority) * capacity / 100.0

            if params['has_bend']:
                # Allow temporary burst up to the full capacity.
                params['ls_m1'] = float(cl.priority) / sum_priority * capacity

                params['ls_d'] = cl.packetdelay
                if cl.packetsize is not None:
                    transfer_delay = float(8 * cl.packetsize) / params['ls_m1']
                    params['ls_d'] = max(params['ls_d'], transfer_delay)

        if params['has_ul']:
            # Set steady state upper limit from a percentage of the capacity.
            params['ul_m2'] = float(cl.limitrate) * capacity / 100.0

    return result


class ConfigInterface(ConfigObject):
    typename = "interface"

    options = [
        ConfigOption(name="enabled", type=bool, required=True),
        ConfigOption(name="classgroup", default="Default"),
        ConfigOption(name="overhead", type=bool, default=True),
        ConfigOption(name="upload", type=int, default=4096),
        ConfigOption(name="download", type=int, default=512)
    ]

    def apply(self, allConfigs):
        commands = list()

        if not self.enabled:
            return commands

        # Look up the interface section.
        interface = self.lookup(allConfigs, "network", "interface", self.name)

        # Store ifname mainly for the revert command.
        self.config_ifname = interface.config_ifname

        classes = []
        default_classid = None
        classgroup = self.lookup(allConfigs, "qos", "classgroup", self.classgroup)
        for class_name, class_id in six.iteritems(classgroup.class_id_map):
            if class_name == classgroup.default:
                default_classid = class_id

            traffic_class = self.lookup(allConfigs, "qos", "class", class_name)
            classes.append((class_id, traffic_class))

        cmd = ["tc", "qdisc", "add",
                "dev", self.config_ifname,
                "root", "handle", "1:",
                "hfsc"]
        if default_classid is not None:
            cmd.extend(["default", str(default_classid)])
        commands.append((self.PRIO_CREATE_QDISC, Command(cmd, self)))

        hfsc_params = compute_hfsc_params(classes, self.upload)
        for class_id, traffic_class in classes:
            full_class_id = "1:{}".format(class_id)
            params = hfsc_params[class_id]

            cmd = [
                "tc", "class", "add",
                "dev", self.config_ifname,
                "parent", "1:",
                "classid", full_class_id,
                "hfsc"
            ]

            if params['has_rt']:
                cmd.append("rt")
                if params['has_bend']:
                    cmd.extend(["m1", str(params['rt_m1']) + "kbit",
                                "d", str(params['rt_d']) + "ms"])
                cmd.extend(["m2", str(params['rt_m2']) + "kbit"])

            if params['has_ls']:
                cmd.append("ls")
                if params['has_bend']:
                    cmd.extend(["m1", str(params['ls_m1']) + "kbit",
                                "d", str(params['ls_d']) + "ms"])
                cmd.extend(["m2", str(params['ls_m2']) + "kbit"])

            if params['has_ul']:
                cmd.extend(["ul", "m2", str(params['ul_m2'])+"kbit"])

            commands.append((self.PRIO_CREATE_QDISC, Command(cmd, self)))

            # Add a fair queue to each class so that flows within the same
            # class receive fair treatment.
            cmd = [
                "tc", "qdisc", "add",
                "dev", self.config_ifname,
                "parent", full_class_id,
                "fq_codel", "noecn"
            ]
            commands.append((self.PRIO_CREATE_QDISC, Command(cmd, self)))

        return commands

    def revert(self, allConfigs):
        commands = list()

        if not self.enabled:
            return commands

        cmd = ["tc", "qdisc", "del", "dev", self.config_ifname, "root"]
        commands.append((-self.PRIO_CREATE_QDISC, Command(cmd, self)))

        return commands


class ConfigClassify(ConfigObject):
    typename = "classify"

    options = [
        ConfigOption(name="target", required=True),
        ConfigOption(name="proto"),
        ConfigOption(name="srchost"),
        ConfigOption(name="dsthost"),
        ConfigOption(name="ports"),
        ConfigOption(name="srcports"),
        ConfigOption(name="dstports"),
        ConfigOption(name="portrange"),
        ConfigOption(name="pktsize"),
        ConfigOption(name="tcpflags"),
        ConfigOption(name="mark"),
        ConfigOption(name="connbytes"),
        ConfigOption(name="tos"),
        ConfigOption(name="dscp"),
        ConfigOption(name="direction")
    ]

    def make_iptables_cmd(self, action, ifname, class_id):
        cmd = [
            "iptables",
            "--wait", "5",
            "--table", "mangle",
            action, "POSTROUTING",
            "--out-interface", ifname
        ]

        if self.proto is not None:
            cmd.extend(["--protocol", self.proto])

        if self.srchost is not None:
            cmd.extend(["--source", self.srchost])
        if self.dsthost is not None:
            cmd.extend(["--destination", self.dsthost])

        if self.ports is not None:
            cmd.extend(["--match", "multiport", "--ports", self.ports])
        if self.srcports is not None:
            cmd.extend(["--match", "multiport", "--source-ports", self.srcports])
        if self.dstports is not None:
            cmd.extend(["--match", "multiport", "--destination-ports", self.dstports])
        if self.portrange is not None:
            cmd.extend(["--match", "multiport", "--ports", self.portrange])

        if self.pktsize is not None:
            cmd.extend(["--match", "connbytes", "--connbytes", self.pktsize])
        if self.tcpflags is not None:
            cmd.extend(["--protocol", "tcp", "--tcp-flags", self.tcpflags])
        if self.mark is not None:
            cmd.extend(["--match", "mark", "--mark", self.mark])
        if self.connbytes is not None:
            cmd.extend(["--match", "connbytes", "--connbytes", self.connbytes])
        if self.tos is not None:
            cmd.extend(["--match", "tos", "--tos", self.tos])
        if self.dscp is not None:
            cmd.extend(["--match", "dscp", "--dscp-class", self.dscp])

        comment = "classify {}".format(self.target)
        cmd.extend([
            "--match", "comment", "--comment", comment,
            "--jump", "CLASSIFY",
            "--set-class", class_id
        ])

        return Command(cmd, self)

    def apply(self, allConfigs):
        self._created_rules = []
        commands = []

        self.lookup(allConfigs, "qos", "class", self.target)

        # Find all classgroup sections that contain the target class.
        for group in self.findByType(allConfigs, "qos", "classgroup"):
            class_id = group.get_class_id(self.target)
            if class_id is None:
                continue
            group.dependents.add(self)

            # Find all interfaces that use the classgroup.
            interface_query = {"classgroup": group.name}
            for interface in self.findByType(allConfigs, "qos", "interface",
                    where=interface_query):
                if not interface.enabled:
                    continue
                interface.dependents.add(self)

                network_interface = self.lookup(allConfigs, "network",
                        "interface", interface.name)

                full_class_id = "1:{}".format(class_id)

                args = (network_interface.config_ifname, full_class_id)
                cmd = self.make_iptables_cmd("--append", *args)

                # Save these arguments so that we can recreate the iptables
                # commands in the revert method.
                self._created_rules.append(args)

                commands.append((self.PRIO_IPTABLES_RULE, cmd))

        return commands

    def revert(self, allConfigs):
        commands = []
        for args in self._created_rules:
            cmd = self.make_iptables_cmd("--delete", *args)
            commands.append((-self.PRIO_IPTABLES_RULE, cmd))
        return commands


class ConfigClassgroup(ConfigObject):
    typename = "classgroup"

    options = [
        ConfigOption(name="classes", required=True),
        ConfigOption(name="default", required=True)
    ]

    def get_class_id(self, class_name):
        """
        Get ID for a traffic class in this group.

        Returns None if the class is not a member of the group.
        """
        return self.class_id_map.get(class_name, None)

    def setup(self):
        # Map names to integer ids.
        self.class_id_map = dict()
        names = self.classes.split(" ")
        for i in range(len(names)):
            self.class_id_map[names[i]] = i+1


class ConfigClass(ConfigObject):
    typename = "class"

    options = [
        ConfigOption(name="packetsize", type=int),
        ConfigOption(name="packetdelay", type=int),
        ConfigOption(name="maxsize", type=int),
        ConfigOption(name="avgrate", type=int),
        ConfigOption(name="limitrate", type=int),
        ConfigOption(name="priority", type=int)
    ]



import json
import os
import time

from paradrop.base import settings
from paradrop.base.output import out
from paradrop.lib.utils.uhttp import UHTTPConnection


SNAP_NAME = os.environ.get('SNAP_NAME', None)


class GovernorClient(object):
    """
    Client for interacting with the governor API for restricted administrative
    functions.
    """
    def __init__(self, logging=True, wait_async=False):
        """
        Initialize client for interacting with governor API.

        logging: whether to log results
        wait_async: whether GovernorClient should wait for async calls to finish
        """
        self.logging = True
        self.wait_async = wait_async

    @classmethod
    def isAvailable(cls):
        """
        Check if governor API is available.
        """
        return os.path.exists(settings.GOVERNOR_INTERFACE)

    def _maybeWait(self, result):
        """
        Wait for completed result if configured to wait on async calls.
        """
        if result['type'] == 'sync' or not self.wait_async:
            return result['result']

        if result['type'] == 'error':
            result = result['result']
            if self.logging:
                out.warn("snapd error: {}".format(result['message']))
            return result

        while True:
            change = self.get_change(result['change'])
            if change['ready']:
                if self.logging:
                    out.info("{}: {}".format(change['summary'], change['status']))
                return change
            time.sleep(1)

    def addAuthorizedKey(self, key, user="paradrop"):
        """
        Add an authorized key.
        """
        conn = UHTTPConnection(settings.GOVERNOR_INTERFACE)
        path = "/users/{}/authorized_keys".format(user)
        body = json.dumps({
            'key': key
        })
        headers = {"Content-type": "application/json"}
        conn.request("POST", path, body, headers)
        res = conn.getresponse()
        return json.loads(res.read().decode('utf-8'))

    def connect(self, plug_snap=SNAP_NAME, plug=None, slot_snap="core", slot=None):
        """
        Connect an interface.
        """
        conn = UHTTPConnection(settings.GOVERNOR_INTERFACE)
        path = "/snapd/v2/interfaces"
        body = json.dumps({
            'action': 'connect',
            'slots': [{'snap': slot_snap, 'slot': slot}],
            'plugs': [{'snap': plug_snap, 'plug': plug}]
        })
        headers = {"Content-type": "application/json"}
        conn.request("POST", path, body, headers)
        res = conn.getresponse()
        data = json.loads(res.read().decode('utf-8'))
        return self._maybeWait(data)

    def get_change(self, change_id):
        """
        Get the current status of a change.
        """
        conn = UHTTPConnection(settings.GOVERNOR_INTERFACE)
        path = "/snapd/v2/changes/{}".format(change_id)
        conn.request("GET", path)
        res = conn.getresponse()
        data = json.loads(res.read().decode('utf-8'))
        return self._maybeWait(data)

    def installSnap(self, snapName):
        """
        Install a snap from the store.
        """
        conn = UHTTPConnection(settings.GOVERNOR_INTERFACE)
        path = "/snapd/v2/snaps"
        body = json.dumps({
            'action': 'install',
            'snaps': [snapName]
        })
        headers = {"Content-type": "application/json"}
        conn.request("POST", path, body, headers)
        res = conn.getresponse()
        data = json.loads(res.read().decode('utf-8'))
        return self._maybeWait(data)

    def listAuthorizedKeys(self, user="paradrop"):
        """
        Get a list of authorized keys.
        """
        path = "/users/{}/authorized_keys".format(user)
        conn = UHTTPConnection(settings.GOVERNOR_INTERFACE)
        conn.request("GET", path)
        res = conn.getresponse()
        return json.loads(res.read().decode('utf-8'))

    def listSnaps(self):
        """
        Get a list of installed snaps.
        """
        conn = UHTTPConnection(settings.GOVERNOR_INTERFACE)
        conn.request("GET", "/snapd/v2/snaps")
        res = conn.getresponse()
        data = json.loads(res.read().decode('utf-8'))
        return self._maybeWait(data)

    def updateSnap(self, snapName, data):
        """
        Post an update to a snap.

        Valid actions are: install, refresh, remove, revert, enable, disable.

        Example:
        updateSnap("paradrop-daemon", {"action": "refresh"}
        """
        conn = UHTTPConnection(settings.GOVERNOR_INTERFACE)
        path = "/snapd/v2/snaps/{}".format(snapName)
        body = json.dumps(data)
        headers = {"Content-type": "application/json"}
        conn.request("POST", path, body, headers)
        res = conn.getresponse()
        data = json.loads(res.read().decode('utf-8'))
        return self._maybeWait(data)



import os

import six

from paradrop.base.output import out
from paradrop.base import pdutils
from paradrop.core.config import uciutils
from paradrop.lib.utils import uci


# Should dnsmasq be used to cache DNS lookup results or should clients send
# their queries directly to specified nameserver(s)?  This question only
# applies if the developer explicitly specified DNS nameservers.  Otherwise,
# dnsmasq acts as a cache and uses whatever nameserver(s) the host
# /etc/resolv.conf file specifies.
DNSMASQ_CACHE_ENABLED = True


def getVirtDHCPSettings(update):
    """
    Looks at the runtime rules the developer defined to see if they want a dhcp
    server.  If so it generates the data and stores it into the chute cache
    key:virtDHCPSettings.
    """

    interfaces = update.cache_get('networkInterfaces')
    if interfaces is None:
        return

    dhcpSettings = list()

    for iface in interfaces:
        # Only look at interfaces with DHCP server requested.
        if 'dhcp' not in iface:
            continue
        dhcp = iface['dhcp']

        # Construct a path for the lease file that will be visible inside the
        # chute.
        leasefile = os.path.join(
            update.cache_get('externalSystemDir'),
            "dnsmasq-{}.leases".format(iface['name'])
        )

        # NOTE: Having one dnsmasq section for each interface deviates from how
        # OpenWRT does things, where they assume a single instance of dnsmasq
        # will be handling all DHCP and DNS needs.
        config = {'type': 'dnsmasq'}
        options = {
            'leasefile': leasefile,
            'interface': [iface['externalIntf']]
        }

        # Optional: developer can pass in a list of DNS nameservers to use
        # instead of the system default.
        #
        # This path -> clients query our dnsmasq server; dnsmasq uses the
        # specified nameservers and caches the results.
        if DNSMASQ_CACHE_ENABLED and 'dns' in dhcp:
            options['noresolv'] = '1'
            options['server'] = dhcp['dns']

        # Disable authoritative mode if serving as a relay.
        relay = dhcp.get('relay', None)
        if relay is not None:
            options['authoritative'] = False

        dhcpSettings.append((config, options))

        # Check for fields that are required if not in relay mode.
        res = pdutils.check(dhcp, dict, ['lease', 'start', 'limit'])
        if relay is None and res:
            out.warn('DHCP server definition {}\n'.format(res))
            raise Exception("DHCP server definition missing field(s)")

        config = {'type': 'dhcp', 'name': iface['externalIntf']}
        options = {
            'interface': iface['externalIntf'],
            'start': dhcp.get('start', None),
            'limit': dhcp.get('limit', None),
            'leasetime': dhcp.get('lease', None),
            'dhcp_option': []
        }

        # This option tells clients that the router is the interface inside the
        # chute not the one in the host.
        options['dhcp_option'].append("option:router,{}".format(iface['internalIpaddr']))

        # Optional: developer can pass in a list of DNS nameservers to use
        # instead of the system default.
        #
        # This path -> clients receive the list of DNS servers and query them
        # directly.
        if not DNSMASQ_CACHE_ENABLED and 'dns' in dhcp:
            options['dhcp_option'].append(",".join(["option:dns-server"] + dhcp['dns']))

        # Interpret the optional DHCP relay configuration.
        # - string: interpret as the IP address of the relay server.
        # - list: interpret as list of strings to pass to dnsmasq (--dhcp-relay options).
        if isinstance(relay, six.string_types):
            # TODO: Set the optional third argument (interface) to the name of
            # the network interface on which we expect DHCP response. This
            # could be the WAN interface or it could be VPN interface.
            options['relay'] = ["{},{}".format(iface['externalIpaddr'], relay)]
        elif isinstance(relay, list):
            options['relay'] = relay

        dhcpSettings.append((config, options))

    update.cache_set('virtDHCPSettings', dhcpSettings)


def setVirtDHCPSettings(update):
    """
    Takes a list of tuples (config, opts) and saves it to the dhcp config file.
    """
    uciutils.setConfig(update,
            cacheKeys=['virtDHCPSettings'],
            filepath=uci.getSystemPath("dhcp"))


def revert_dhcp_settings(update):
    uciutils.setConfig(update,
            cacheKeys=['virtDHCPSettings'],
            filepath=uci.getSystemPath("dhcp"))



###################################################################
# Copyright 2013-2014 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

import os
import pickle
from twisted.internet.task import LoopingCall

from paradrop.base.output import out
from paradrop.lib.utils import pdos
from paradrop.lib.utils.yaml import yaml


class PDStorage(object):

    """
        ParaDropStorage class.

        This class is designed to be implemented by other classes.
        Its purpose is to make whatever data is considered important persistant to disk.

        The implementer can override functions in order to implement this class:
            getAttr() : Get the attr we need to save to disk
            setAttr() : Set the attr we got from disk
            importAttr(): Takes a payload and returns the properly formatted data
            exportAttr(): Takes the data and returns a payload
            attrSaveable(): Returns True if we should save this attr
    """

    def __init__(self, filename, saveTimer):
        self.filename = filename
        self.saveTimer = saveTimer

        # Setup looping call to keep chute list perisistant
        if (self.saveTimer > 0):
            self.repeater = LoopingCall(self.saveToDisk)
            self.repeater.start(self.saveTimer)

    def loadFromDisk(self):
        """Attempts to load the data from disk.
            Returns True if success, False otherwise."""

        if(pdos.exists(self.filename)):
            deleteFile = False
            try:
                pyld = pickle.load(pdos.open(self.filename, 'rb'))
                self.setAttr(self.importAttr(pyld))
                return True
            except Exception as e:
                out.err('Error loading from disk: %s\n' % (str(e)))
                deleteFile = True

            # Delete the file
            if(deleteFile):
                try:
                    pdos.unlink(self.filename)
                except Exception as e:
                    out.err('Error unlinking %s\n' % (self.filename))

        return False

    def saveToDisk(self):
        """Saves the data to disk."""
        out.info('Saving to disk (%s)\n' % (self.filename))

        # Make sure they want to save
        if(not self.attrSaveable()):
            return

        # Get whatever the data is
        pyld = self.exportAttr(self.getAttr())

        # Write the file to disk, truncate if it exists
        try:
            with open(self.filename, 'wb') as output:
                pickle.dump(pyld, output)
                os.fsync(output.fileno())
        except Exception as e:
            out.err('Error writing to disk %s\n' % (str(e)))

        try:
            with open(self.filename + ".yaml", "w") as output:
                yaml.dump(pyld, output)
        except Exception as error:
            out.err("Error writing yaml file: {}".format(error))

    def attrSaveable(self):
        """THIS SHOULD BE OVERRIDEN BY THE IMPLEMENTER."""
        return False

    def importAttr(self, pyld):
        """By default do nothing, but expect that this function could be overwritten"""
        return pyld

    def exportAttr(self, data):
        """By default do nothing, but expect that this function could be overwritten"""
        return data



import base64
import functools
import json

from klein import Klein
from twisted.internet import defer

from paradrop.base import nexus, settings
from paradrop.base.output import out
from paradrop.core.agent.http import PDServerRequest
from paradrop.core.auth.user import User
from paradrop.core.chute.chute_storage import ChuteStorage
from . import cors

# TODO: Configurable username/password.
# TODO: Segment permissions based on roles.
#       Admin can do anything, but maybe a guest can still view device status.
#       Chutes should not be able to manipulate other chutes.


def get_username_password(userpass):
    """
    Please note: username and password can either be presented in plain text
    such as "admin:password" or base64 encoded such as "YWRtaW46cGFzc3dvcmQ=".
    Both forms should be returned from this function.
    """
    plaintext = base64.b64decode(userpass)

    segments = plaintext.split(':')
    count = len(segments)
    if count == 1:
        return segments[0], ''
    elif count == 2:
        return segments
    else:
        raise Exception('There should be at most two segments!')


def verify_password(password_manager, userpass):
    user_name, password = get_username_password(userpass)
    return password_manager.verify_password(user_name, password)


def get_allowed_bearer():
    """
    Return set of allowed bearer tokens.
    """
    allowed = set()

    chuteStore = ChuteStorage()
    chutes = chuteStore.getChuteList()
    for chute in chutes:
        token = chute.getCache('apiToken')
        if token is not None:
            allowed.add(token)

    return allowed


def check_auth(request, password_manager, token_manager):
    auth_header = request.getHeader('Authorization')
    if auth_header is None:
        return False

    parts = auth_header.split()
    if parts[0] == "Basic":
        username, password = get_username_password(parts[1])
        request.user = User(username, "localhost", role="admin")
        return password_manager.verify_password(username, password)
    elif parts[0] == "Bearer":
        token = parts[1]

        # Chutes use non-expiring random tokens that are generated at container
        # creation. This works well because we do not want to deal with
        # expiration or revocation.
        allowed_tokens = get_allowed_bearer()
        if token in allowed_tokens:
            return True

        # Users (through the local portal or pdtools) can acquire expiring
        # JWTs. If the JWT decodes using our secret, then the caller is
        # authenticated.
        #
        # Later on, we may implement tokens issued by paradrop.org but
        # allowing user access to the router, etc. In that case we will
        # need to examine the subject, issuer, and audience claims.
        try:
            data = token_manager.decode(token)
            domain = data.get("domain", "localhost")
            username = data.get("sub", "paradrop")
            role = data.get("role", "user")
            request.user = User(username, domain, role=role)
            return True
        except token_manager.InvalidTokenError:
            pass

    return False


def requires_auth(func):
    """
    Use as a decorator for API functions to require authorization.

    This checks the Authorization HTTP header.  It handles username and
    password as well as bearer tokens.
    """
    @functools.wraps(func)
    def decorated(self, request, *args, **kwargs):
        if not check_auth(request, self.password_manager, self.token_manager):
            out.info('HTTP {} {} {} {}'.format(request.getClientIP(),
                request.method, request.path, 401))
            request.setResponseCode(401)
            request.setHeader("WWW-Authenticate", "Basic realm=\"Login Required\"")
            return
        return func(self, request, *args, **kwargs)
    return decorated


def verify_cloud_token(token):
    headers = {
        "Authorization": "Bearer {}".format(token)
    }

    # Pass custom Authorization header and setAuthHeader=False to prevent
    # PDServerRequest from using the node's own authorization token.
    request = PDServerRequest("/api/users/me", headers=headers,
            setAuthHeader=False)
    return request.get()


def get_access_level(user, node):
    access_rights = node.get('access_rights', [])

    roles = ["guest"]
    for item in access_rights:
        if item['user_id'] == user['_id']:
            roles.append(item['role'])

    return User.get_highest_role(roles)


class AuthApi(object):
    routes = Klein()

    def __init__(self, password_manager, token_manager):
        self.password_manager = password_manager
        self.token_manager = token_manager

    @routes.route('/local', methods=['POST'])
    def local_login(self, request):
        """
        Login using local authentication (username+password).
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        body = json.loads(request.content.read())

        # If username is missing in the body, automatically fill in the
        # default.  If password is missing, assume empty string regardless of
        # the default.  The caller can only login without a password if the
        # current password is empty.
        username = body.get('username', settings.DEFAULT_PANEL_USERNAME)
        password = body.get('password', '')

        success = self.password_manager.verify_password(username, password)
        result = {
            'username': username,
            'success': success
        }

        if success:
            token = self.token_manager.issue(username, domain="localhost",
                    role="admin")
            result['token'] = token
        else:
            request.setResponseCode(401)

        return json.dumps(result)

    @routes.route('/cloud', methods=['POST'])
    def auth_cloud(self, request):
        """
        Login using credentials from the cloud controller.

        This is an experimental new login method that lets users
        present a token that they received from the cloud controller
        as a login credential for a node. The idea is to enable easy
        access for multiple developers to share a node, for example,
        during a tutorial.

        Instead of a username/password, the user presents a token received
        from the cloud controller. The verify_cloud_token function
        verifies the validity of the token with the controller, and if
        successful, retrieves information about the bearer, particularly
        the username and role. Finally, we generate a new token that
        enables the user to authenticate with local API endpoints.
        """
        cors.config_cors(request)
        request.setHeader('Content-Type', 'application/json')

        body = json.loads(request.content.read())
        token = body.get('token', '')

        headers = {
            "Authorization": "Bearer {}".format(token)
        }
        node_id = nexus.core.info.pdid

        # Pass custom Authorization header and setAuthHeader=False to prevent
        # PDServerRequest from using the node's own authorization token.
        d1 = PDServerRequest("/api/users/me",
                headers=headers,
                setAuthHeader=False
            ).get()
        d2 = PDServerRequest("/api/routers/{}".format(node_id),
                headers=headers,
                setAuthHeader=False
            ).get()

        def token_verified(responses):
            for response in responses:
                if not response.success or response.data is None:
                    request.setResponseCode(401)
                    return

            remote_user = responses[0].data
            node_info = responses[1].data

            role = get_access_level(remote_user, node_info)

            token = self.token_manager.issue(remote_user['email'],
                    domain="paradrop.org", role=role)

            result = {
                'token': token,
                'username': remote_user['email']
            }
            return json.dumps(result)

        d = defer.gatherResults([d1, d2])
        d.addCallback(token_verified)
        return d



import heapq
import json
import os
import threading

import six

from paradrop.base.output import out
from paradrop.lib.utils import pdosq
from paradrop.lib.utils.uci import UCIConfig, getSystemConfigDir

# Import all of the modules defining section types, so that all subclasses of
# ConfigObject are known. These are imported only for their side effects.
from . import dhcp
from . import firewall
from . import network
from . import parprouted
from . import qos
from . import wireless

from .base import ConfigObject
from .command import CommandList, ErrorCommand


# Silence pyflakes warning about unused imports.
assert dhcp
assert firewall
assert network
assert parprouted
assert qos
assert wireless


# Map of type names to the classes that handle them.  We now prefer the
# extended type name, e.g. "network:interface", because there can be
# conflicting types, e.g. "qos:interface".
configTypeMap = dict()
for cls in ConfigObject.__subclasses__():
    extended = "{}:{}".format(cls.getModule(), cls.typename)
    configTypeMap[extended] = cls

    existing = configTypeMap.get(cls.typename, None)
    if existing is None or existing.maskable:
        configTypeMap[cls.typename] = cls


def findConfigFiles(search=None):
    """
    Look for and return a list of configuration files.

    The behavior depends on whether the search argument is a file, a directory,
    or None.

    If search is None, return a list of files in the system config directory.
    If search is a file name (not a path), look for it in the working directory
    first, and the system directory second.  If search is a full path to a
    file, and it exists, then return that file.  If search is a directory,
    return the files in that directory.
    """
    if search is None:
        search = getSystemConfigDir()

    files = list()
    if os.path.isfile(search):
        files.append(search)
    elif os.path.isdir(search):
        for fn in os.listdir(search):
            path = os.path.join(search, fn)
            files.append(path)
    else:
        path = os.path.join(getSystemConfigDir(), search)
        if os.path.isfile(path):
            files.append(path)

    return files


class ConfigManager(object):

    def __init__(self, writeDir, execCommands=True):
        """
        writeDir: directory to use for generated config files (e.g. hostapd.conf).
        execCommands: whether or not to run commands (set to False for testing).
        """
        self.writeDir = writeDir
        self.execCommands = execCommands

        # Make sure directory exists.
        pdosq.makedirs(writeDir)

        self.previousCommands = list()
        self.currentConfig = dict()
        self.nextSectionId = 0

        # Number of objects requiring IP forwarding.
        # If >0, we need to enable system-wide.
        # If ==0, we can probably disable.
        self.forwardingCount = 0

        # Track whether we have loaded the conntrack kernel module.  We need to
        # make sure the module is loaded before adding iptables rules related
        # to connection state.
        self.conntrackLoaded = False

        # Allow threads to wait for first load to complete.  This will be set
        # after the first load completes and will remain set thereafter.
        self.systemUp = threading.Event()

        # Track epoch number so that we know which configuration sections
        # were applied in the most recent epoch.
        self.epoch = 0

    def changingSet(self, files):
        """
        Return the sections from the current configuration that may have
        changed.

        This checks which sections from the current configuration came from
        files in the given file list.  These are sections that may be changed
        or removed when we reload the files.
        """
        out = set()
        files = set(files)
        for config in self.currentConfig.values():
            if config.source in files:
                out.add(config)
        return out

    def getPreviousCommands(self):
        """
        Get the most recent command list.
        """
        return self.previousCommands

    def execute(self, commands):
        """
        Execute commands.

        Takes a CommandList object.
        """
        for cmd in commands.commands():
            cmd.execute()

    def findMatchingConfig(self, config, byName=False):
        """
        Check the current config for an identical section.

        Returns the matching object or None.
        """
        # First try by name (faster).
        key = config.getTypeAndName()

        if key in self.currentConfig:
            oldConfig = self.currentConfig[key]
            if byName:
                return oldConfig
            if config.optionsMatch(oldConfig):
                return oldConfig

        # Loop over them and check by content.
        for oldConfig in self.currentConfig.values():
            if config.optionsMatch(oldConfig):
                return oldConfig

        return None

    def loadConfig(self, search=None, execute=True):
        """
        Load configuration files and apply changes to the system.

        We process the configuration files in sections.  Each section
        corresponds to an interface, firewall rule, DHCP server instance, etc.
        Each time we reload configuration files after the initial time, we
        check for changes against the current configuration.  Here is the
        decision tree for handling differences in the newly loaded
        configuration vs. the existing configuration:

        Section exists in current config (by type and name)?
         - No -> Add section, apply changes, and stop.
         - Yes -> Continue.
        Section is identical to the one in the current config (by option
        values)?
         - No -> Revert current section, mark any affected dependents,
                 add new section, apply changes, and stop.
         - Yes -> Continue.
        Section has not changed but one of its dependencies has?
         - No -> Stop.
         - Yes -> Revert current section, mark any affected dependents,
                  add new section, apply changes, and stop.
        """
        self.epoch += 1

        # Map (type, name) -> config
        allConfigs = dict(self.currentConfig)

        # Manage sets of configuration sections.
        # newConfigs: completely new or new versions of existing sections.
        # affectedConfigs: sections that are affected due to dependency
        # changing.
        # undoConfigs: old sections that need to be undone before proceeding.
        newConfigs = set()
        updatedConfigs = set()
        undoConfigs = set()

        # Final list of commands to execute.
        commands = CommandList()

        files = findConfigFiles(search)

        # We will remove things from this set as we find them in the new
        # configuration files.  Anything that remains at the end must have been
        # deleted from its source file.
        deletionSet = self.changingSet(files)

        for config in self.readConfig(files):
            # Check if the section already exists in identical form
            # in our current configuration.
            matchByContent = self.findMatchingConfig(config, byName=False)
            matchByName = self.findMatchingConfig(config, byName=True)

            # If we found anything that matches, then the existing config
            # section should not be deleted.
            if matchByContent in deletionSet:
                deletionSet.remove(matchByContent)
            if matchByName in deletionSet:
                deletionSet.remove(matchByName)

            # If an existing section is identical (in content), we have
            # no new work to do for this section.
            if matchByContent is not None:
                # Just make sure the new config's internalName matches the old
                # config's.  This is mainly for anonymous sections because they
                # have auto-generated unique names.
                config.internalName = matchByContent.internalName

            else:
                if matchByName is not None:
                    # Old section will need to be updated appropriately.
                    updatedConfigs.add(matchByName)

                    # Keep track of sections that may be affected by this
                    # one's change.
                    updatedConfigs.update(matchByName.dependents)

                    # Remove links to the old version because it will be
                    # redundant after the new configuration goes into effect.
                    matchByName.removeFromParents()

                else:
                    # If it did not exist or is different, add it to our queue
                    # of sections to execute.
                    newConfigs.add(config)

                # Mark that the config section was applied in this epoch.
                config.epoch = self.epoch

                allConfigs[config.getTypeAndName()] = config

        # Items from the deletion set should be deleted from memory as well as
        # have their changes undone.
        for config in deletionSet:
            del allConfigs[config.getTypeAndName()]
            config.removeFromParents()
            undoConfigs.add(config)

        # Remove configs that are in both sets---we should not try to reload
        # updated configs that are supposed to be removed.
        updatedConfigs -= undoConfigs

        undoWork = ConfigObject.prioritizeConfigs(updatedConfigs | undoConfigs,
                reverse=True)
        heapq.heapify(undoWork)

        while undoWork:
            prio, config = heapq.heappop(undoWork)

            try:
                if config in undoConfigs:
                    commands.extend(config.revert(allConfigs))
                else:
                    new = allConfigs[config.getTypeAndName()]
                    commands.extend(config.updateRevert(new, allConfigs))
            except Exception as error:
                commands.append(prio, ErrorCommand(error, config))

        applyWork = ConfigObject.prioritizeConfigs(updatedConfigs | newConfigs)
        heapq.heapify(applyWork)

        while applyWork:
            prio, config = heapq.heappop(applyWork)

            try:
                if config in newConfigs:
                    commands.extend(config.apply(allConfigs))
                else:
                    new = allConfigs[config.getTypeAndName()]
                    commands.extend(config.updateApply(new, allConfigs))
            except Exception as error:
                commands.append(prio, ErrorCommand(error, config))

            # Reset list of executed commands so that previous failures do not
            # haunt us.
            config.executed = list()

        # Finally, execute the commands.
        if execute and self.execCommands:
            self.execute(commands)

        self.previousCommands = commands
        self.currentConfig = allConfigs

        # Wake up anything that was waiting for the first load to complete.
        self.systemUp.set()

        return self.statusString()

    def readConfig(self, files):
        """
        Load configuration files and return configuration objects.

        This method only loads the configuration files without making any
        changes to the system and returns configuration objects as a generator.
        """
        # Keep track of headers (section type and name) that have been
        # processed so far.  The dictionary maps them to filename, so that we
        # can print a useful warning message about duplicates.
        usedHeaders = dict()

        for fn in files:
            out.info("Reading file {}\n".format(fn))

            # Extract just the filename (e.g. wireless, network, qos).
            basename = os.path.basename(fn)

            uci = UCIConfig(fn)
            config = uci.readConfig()

            for section, options in config:
                # Sections differ in where they put the name, if they have one.
                if "name" in section:
                    name = section['name']
                elif "name" in options:
                    name = options['name']
                else:
                    name = None

                # Get section comment string (optional, but Paradop uses it).
                comment = section.get('comment', None)

                # Try looking up by extended type first.  It turns out, both
                # "network" and "qos" configuration files should support
                # sections with typename "interface".  We can disambiguate
                # them by calling the latter one "qos:interface".
                ext_type = "{}:{}".format(basename, section['type'])
                if ext_type in configTypeMap:
                    cls = configTypeMap[ext_type]
                elif section['type'] in configTypeMap:
                    cls = configTypeMap[section['type']]
                else:
                    out.warn("Unsupported section type {} in {}\n".format(
                        section['type'], fn))
                    continue

                try:
                    obj = cls.build(self, fn, name, options, comment)
                except Exception as e:
                    out.warn("Error building object from section {}:{} in "
                             "{}\n".format(section['type'], name, fn))
                    out.warn(e)
                    continue

                key = obj.getTypeAndName()
                if key in usedHeaders:
                    out.warn("Section {}:{} from {} overrides section in "
                             "{}\n".format(section['type'], name, fn,
                                           usedHeaders[key]))
                usedHeaders[key] = fn

                yield obj

    def statusString(self):
        """
        Return a JSON string representing status of the system.

        The format will be a list of dictionaries.  Each dictionary
        corresponds to a configuration block and contains at the following
        fields.

        type: interface, wifi-device, etc.
        name: name of the section (may be autogenerated for some configs)
        comment: comment from the configuration file or None
        success: True if all setup commands succeeded
        """
        status = list()
        for key, config in six.iteritems(self.currentConfig):
            success = all(cmd.success() for cmd in config.executed)
            configStatus = {
                'type': config.typename,
                'name': config.name,
                'comment': config.comment,
                'success': success,
                'age': self.epoch - config.epoch
            }
            status.append(configStatus)
        return json.dumps(status)

    def unload(self, execute=True):
        commands = CommandList()

        undoWork = ConfigObject.prioritizeConfigs(self.currentConfig.values())
        heapq.heapify(undoWork)

        while undoWork:
            prio, config = heapq.heappop(undoWork)
            commands.extend(config.revert(self.currentConfig))

        # Finally, execute the commands.
        if execute and self.execCommands:
            self.execute(commands)

        self.previousCommands = commands
        self.currentConfig = dict()
        return True

    def waitSystemUp(self):
        """
        Wait for the first load to complete and return system status string.
        """
        self.systemUp.wait()
        return self.statusString()



import os

from paradrop.core.config import devices


class AirsharkInterfaceManager(object):
    def __init__(self):
        self.interface = None
        self.observers = []

    def interface_available(self):
        return self.interface != None

    def add_observer(self, observer):
        if self.observers.count(observer) == 0:
            self.observers.append(observer)

    def remove_observer(self, observer):
        if self.observers.count(observer) == 1:
            self.observers.remove(observer)

    def set_interface(self, interface):
        self.reset_interface()

        os.system("ip link set %s down" % interface)
        os.system("iw dev %s set type managed" % interface)
        os.system("ip link set %s up" % interface)

        self.interface = interface
        for observer in self.observers:
            observer.on_interface_up(self.interface)

    def reset_interface(self):
        if (self.interface is not None):
            for observer in self.observers:
                observer.on_interface_down(self.interface)

            self.interface = None

airshark_interface_manager = AirsharkInterfaceManager()


def configure(update):
    """
    Configure an Airshark interface.
    """
    hostConfig = update.cache_get('hostConfig')
    networkDevices = update.cache_get('networkDevices')

    interfaces = hostConfig.get('wifi-interfaces', [])
    airshark_interfaces = []

    # Look for objects in wifi-interfaces with mode == 'airshark'.
    for interface in interfaces:
        if interface.get('mode', None) != 'airshark':
            continue

        device = devices.resolveWirelessDevRef(interface['device'],
                networkDevices)
        if device['primary_interface'] is not None:
            airshark_interfaces.append(device['primary_interface'])

    if len(airshark_interfaces) > 1:
        raise Exception("Only one wifi interface can be in airshark mode")
    elif len(airshark_interfaces) == 1:
        airshark_interface = airshark_interfaces[0]
        airshark_interface_manager.set_interface(airshark_interface)
    else:
        airshark_interface_manager.reset_interface()



from paradrop.lib.utils import uci


def setConfig(update, cacheKeys, filepath):
    """
    Helper function used to modify config file of each various setting in /etc/config/
    Returns:
        True: if it modified a file
        False: if it did NOT cause any modifications
    Raises exception if an error occurs.
    """
    # First pull out all the cache keys from the @new chute
    newconfigs = []

    # chute may be None if a chute installation failed and we are backing out.
    # In that case, newconfigs should be an empty list, meaning there should be
    # nothing left of the chute in the configuration file after we are done.
    if update.new is not None:
        for c in cacheKeys:
            t = update.cache_get(c)
            if t is not None:
                newconfigs.extend(t)

    # Get the chute name. At least one of chute or old should be valid.
    chute_name = update.new.name if update.new is not None else update.old.name

    # Add comment to each config so we can differentiate between different
    # chute specific configs.
    for c, o in newconfigs:
        c['comment'] = chute_name

    # Get the old configs from the file for this chute.
    cfgFile = uci.UCIConfig(filepath)

    # Get all the configs that existed in the old version. Note we are getting
    # the old configs from the etc/config/ file instead of the chute object.
    # This is to improve reliability  - sometimes the file isn't changed it
    # should be because we have reset the board, messed with chutes, etc. and
    # the new/old chuteobjects are identical.
    oldconfigs = cfgFile.getChuteConfigs(chute_name)

    if (uci.chuteConfigsMatch(oldconfigs, newconfigs)):
        # configs match, skipping reloading
        # Save a backup in case we need to restore.
        cfgFile.backup(backupToken="paradrop")
        return False
    else:
        # We need to make changes so delete old configs, load new configs
        # configs don't match, changing chutes and reloading
        cfgFile.delConfigs(oldconfigs)
        cfgFile.addConfigs(newconfigs)
        cfgFile.save(backupToken="paradrop", internalid=chute_name)
        return True


def restoreConfigFile(chute, configname):
    """
    Restore a system config file from backup.

    This can only be used during a chute update operation to revert changes
    that were made during that update operation.

    configname: name of configuration file ("network", "wireless", etc.)
    """
    filepath = uci.getSystemPath(configname)
    cfgFile = uci.UCIConfig(filepath)
    cfgFile.restore(backupToken="paradrop", saveBackup=False)



import json
import socket


SOCKET_ADDRESS = "/var/run/pdinstall.sock"


def sendCommand(command, data):
    """
    Send a command to the pdinstall service.

    Commands:
    install - Install snaps from a file path or http(s) URL.
        Required data fields:
        sources - List with at least one snap file path or URL.  The snaps
                  are installed in order until one succeeds or all fail.

    Returns True/False for success.  Currently, we cannot check whether the
    call succeeded, only whether it was delived.  A return value of False means
    we could not deliver the command to pdinstall.
    """
    data['command'] = command

    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    try:
        sock.connect(SOCKET_ADDRESS)
        sock.send(json.dumps(data))
        return True
    except:
        return False
    finally:
        sock.close()



import re


class Modalias(object):
    """
    Modalias strings might be a useful way for us to identify hardware.

    http://people.skolelinux.org/pere/blog/Modalias_strings___a_practical_way_to_map__stuff__to_hardware.html
    """
    COMPONENTS = re.compile(r"([a-z]+)([0-9A-F]+)")
    NAMES = {
        "v": "vendor",
        "p": "product",
        "d": "device",
        "sv": "subvendor",
        "sd": "subdevice",
        "bc": "bus_class",
        "sc": "bus_subclass",
        "i": "interface",
        "ic": "interface_class"
    }

    def __init__(self, data):
        self.modalias = data
        self.subtype, contents = data.split(":")

        self.vendor = None
        self.product = None
        self.device = None
        self.subvendor = None
        self.subdevice = None
        self.bus_class = None
        self.bus_subclass = None
        self.interface = None
        self.interface_class = None

        for label, value in Modalias.COMPONENTS.findall(contents):
            if label in Modalias.NAMES:
                setattr(self, Modalias.NAMES[label], value)

    def __repr__(self):
        return self.modalias

    def __str__(self):
        return self.modalias



import json

from autobahn.twisted.websocket import WebSocketServerProtocol
from autobahn.twisted.websocket import WebSocketServerFactory

from paradrop.base.output import out


class StatusSockJSProtocol(WebSocketServerProtocol):
    def __init__(self, factory):
        WebSocketServerProtocol.__init__(self)
        self.factory = factory

    def onOpen(self):
        out.info('sockjs /status connected')

    def onMessage(self, data, isBinary):
        if (data == 'refresh'):
            status = self.factory.system_status.getStatus()
            self.sendMessage(json.dumps(status))

    def onClose(self, wasClean, code, reason):
        out.info('sockjs /status disconnected')


class StatusSockJSFactory(WebSocketServerFactory):
    def __init__(self, system_status):
        WebSocketServerFactory.__init__(self)
        self.system_status = system_status

    def buildProtocol(self, addr):
        return StatusSockJSProtocol(self)



###################################################################
# Copyright 2013-2015 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

"""
Functions associated with deploying and cleaning up docker containers.
"""

import docker
import json
import os
import random
import re
import subprocess
import time

import six

from twisted.internet.threads import deferToThread

from paradrop.base.exceptions import ChuteNotFound
from paradrop.base.output import out
from paradrop.base import constants, nexus, settings
from paradrop.core.config.devices import resetWirelessDevice

from .chutecontainer import ChuteContainer
from .dockerfile import Dockerfile


DOCKER_CONF = """
# Docker systemd configuration
#
# This configuration file was automatically generated by Paradrop.  Any changes
# will be overwritten on startup.

# Tell docker not to start containers automatically on startup.
DOCKER_OPTIONS="--restart=false"
"""

# Used to match and suppress noisy progress messages from Docker output.
#
# Example:
# Extracting
# 862a3e9af0ae
# [================================================>  ] 64.06 MB/65.7 MB
suppress_re = re.compile("^(Downloading|Extracting|[a-z0-9]+|\[=*>?\s*\].*)$")


def getPortList(chute):
    """
    Get a list of ports to expose in the format expected by create_container.

    Uses the port binding dictionary from the chute host_config section.
    The keys are expected to be integers or strings in one of the
    following formats: "port" or "port/protocol".

    Example:
    port_bindings = {
        "1111/udp": 1111,
        "2222": 2222
    }
    getPortList returns [(1111, 'udp'), (2222, 'tcp')]
    """
    if not hasattr(chute, 'host_config') or chute.host_config == None:
        config = {}
    else:
        config = chute.host_config

    ports = []
    for port in config.get('port_bindings', {}).keys():
        if isinstance(port, int):
            ports.append((port, 'tcp'))
            continue

        parts = port.split('/')
        if len(parts) == 1:
            ports.append((int(parts[0]), 'tcp'))
        else:
            ports.append((int(parts[0]), parts[1]))

    # If the chute is configured to host a web service, check
    # whether there is already an entry in the list for the
    # web port.  If not, we should add one.
    web_port = chute.getWebPort()
    if web_port is not None:
        if not any(p[0] == web_port for p in ports):
            ports.append((web_port, 'tcp'))

    return ports


def writeDockerConfig():
    """
    Write options to Docker configuration.

    Mainly, we want to tell Docker not to start containers automatically on
    system boot.
    """
    # First we have to find the configuration file.
    # On ubuntu 16.04 with docker snap, it should be in
    # "/var/snap/docker/{version}/etc/docker/", but version could change.
    path = "/var/snap/docker/current/etc/docker/docker.conf"

    written = False
    if os.path.exists(path):
        try:
            with open(path, "w") as output:
                output.write(DOCKER_CONF)
            written = True
        except Exception as e:
            out.warn('Error writing to {}: {}'.format(path, str(e)))

    if not written:
        out.warn('Could not write docker configuration.')
    return written


def prepare_image(update, service):
    """
    Prepare a Docker image for execution.

    This is usually the longest operation during a chute installation, so
    instead of running this step in the update thread, we spin off a worker
    thread and return a Deferred. This will suspend processing of the current
    update until the worker thread finishes.
    """
    client = docker.APIClient(base_url="unix://var/run/docker.sock", version='auto')

    image_name = service.get_image_name()

    def call(func, *args, **kwargs):
        return func(*args, **kwargs)

    if settings.CONCURRENT_BUILDS:
        wrapper = deferToThread
    else:
        wrapper = call

    if service.type == "image":
        return wrapper(_pull_image, update, client, image_name)

    elif service.type == "inline":
        return wrapper(_build_image, update, service, client, True,
                rm=True, tag=image_name, fileobj=service.dockerfile)

    else:
        return wrapper(_build_image, update, service, client, False,
                rm=True, tag=image_name, path=update.workdir)


def check_image(update, service):
    """
    Check if image exists.
    """
    image_name = service.get_image_name()

    client = docker.DockerClient(base_url="unix://var/run/docker.sock",
            version='auto')

    # Raises an exception if the image does not exist.
    client.images.get(image_name)


def remove_image(update, service):
    """
    Remove a Docker image.
    """
    client = docker.APIClient(base_url="unix://var/run/docker.sock", version='auto')

    image_name = service.get_image_name()
    out.info("Removing image {}\n".format(image_name))

    try:
        client = docker.DockerClient(base_url="unix://var/run/docker.sock",
                version='auto')
        client.images.remove(image=image_name)
    except Exception as error:
        out.warn("Error removing image: {}".format(error))


def create_bridge(update):
    """
    Create a user-defined bridge network for the chute.
    """
    client = docker.DockerClient(base_url="unix://var/run/docker.sock", version='auto')
    client.networks.create(update.new.name, driver="bridge")


def remove_bridge(update):
    """
    Remove the bridge network associated with the chute.
    """
    client = docker.DockerClient(base_url="unix://var/run/docker.sock", version='auto')
    try:
        network = client.networks.get(update.new.name)
        network.remove()
    except docker.errors.NotFound:
        pass


def start_container(update, service):
    """
    Start running a service in a new container.
    """
    client = docker.DockerClient(base_url="unix://var/run/docker.sock", version='auto')

    container_name = service.get_container_name()
    image_name = service.get_image_name()
    out.info("Attempting to start container {} from image {}\n".format(
        container_name, image_name))

    host_config = build_host_config(update, service)

    # TODO
    # Set environment variables for the new container.
    # PARADROP_ROUTER_ID can be used to change application behavior based on
    # what router it is running on.
    environment = prepare_environment(update, service)

    try:
        container = client.containers.run(detach=True, image=image_name,
                name=container_name, environment=environment, **host_config)
        out.info("Successfully started chute with Id: %s\n" % (str(container.id)))
    except Exception as e:
        raise e

    try:
        network = client.networks.get(update.new.name)
        network.connect(container_name, aliases=[service.name])
    except docker.errors.NotFound:
        out.warn("Bridge network {} not found; connectivity between containers is limited.".format(update.new.name))


def remove_container(update, service):
    """
    Remove a service's container.
    """
    container_name = service.get_container_name()
    out.info("Removing container {}\n".format(container_name))

    try:
        client = docker.DockerClient(base_url="unix://var/run/docker.sock",
                version='auto')

        # Grab the last 40 log messages to help with debugging.
        container = client.containers.get(container_name)
        logs = container.logs(stream=False, tail=40, timestamps=False)
        update.progress("{}: {}".format(container_name, logs.rstrip()))

        container.remove(force=True)
    except Exception as error:
        out.warn("Error removing container: {}".format(error))


def _build_image(update, service, client, inline, **buildArgs):
    """
    Build the Docker image and monitor progress (worker function).

    inline: whether Dockerfile is specified as a string or a file in the
    working path.
    """
    # If this is a light chute, generate a Dockerfile.
    if service.type == "light":
        buildArgs['pull'] = True

        dockerfile = Dockerfile(service)
        valid, reason = dockerfile.isValid()
        if not valid:
            raise Exception("Invalid configuration: {}".format(reason))

        if inline:
            # Pass the dockerfile string directly.
            buildArgs['fileobj'] = dockerfile.getBytesIO()
        else:
            # Write it out to a file in the working directory.
            path = os.path.join(buildArgs['path'], "Dockerfile")
            try:
                dockerfile.writeFile(path)
            except Exception as error:
                update.progress(str(error))

    output = client.build(**buildArgs)

    buildSuccess = True
    for line in output:
        #if we encountered an error make note of it
        if 'errorDetail' in line:
            buildSuccess = False

        for key, value in six.iteritems(json.loads(line)):
            if isinstance(value, dict):
                continue
            else:
                msg = value.rstrip()
                if len(msg) > 0 and suppress_re.match(msg) is None:
                    update.progress(msg)

    if not buildSuccess:
        raise Exception("Error building Docker image")


def _pull_image(update, client, image_name):
    """
    Pull the image from a registry.

    Returns True on success, False on failure.
    """
    #auth_config = {
    #    'username': settings.REGISTRY_USERNAME,
    #    'password': settings.REGISTRY_PASSWORD
    #}

    update.progress("Pulling image: {}".format(image_name))

    layers = 0
    complete = 0

    output = client.pull(image_name, auth_config=None, stream=True)
    for line in output:
        data = json.loads(line)

        # Suppress lines that have progressDetail set.  Those are the ones with
        # the moving progress bar.
        if data.get('progressDetail', {}) == {}:
            if 'status' not in data or 'id' not in data:
                continue

            update.progress("{}: {}".format(data['status'], data['id']))

            # Count the number of layers that need to be pulled and the number
            # completed.
            status = data['status'].strip().lower()
            if status == 'pulling fs layer':
                layers += 1
            elif status == 'pull complete':
                complete += 1

    update.progress("Finished pulling {} / {} layers".format(complete, layers))
    if complete < layers:
        raise Exception("Error pulling Docker image")


def stopChute(update):
    """
    Stop a docker container based on the passed in update.

    :param update: The update object containing information about the chute.
    :type update: obj
    :returns: None
    """
    out.info('Attempting to stop chute %s\n' % (update.name))

    c = docker.DockerClient(base_url='unix://var/run/docker.sock', version='auto')
    container = c.containers.get(update.name)
    container.stop()


def restartChute(update):
    """
    Start a docker container based on the passed in update.

    :param update: The update object containing information about the chute.
    :type update: obj
    :returns: None
    """
    out.info('Attempting to restart chute %s\n' % (update.name))
    c = docker.DockerClient(base_url='unix://var/run/docker.sock', version='auto')
    container = c.containers.get(update.name)
    container.start()


def getBridgeGateway():
    """
    Look up the gateway IP address for the docker bridge network.

    This is the docker0 IP address; it is the IP address of the host from the
    chute's perspective.
    """
    client = docker.DockerClient(base_url="unix://var/run/docker.sock", version='auto')

    network = client.networks.get("bridge")
    for config in network.attrs['IPAM']['Config']:
        if 'Gateway' in config:
            return config['Gateway']

    # Fall back to a default if we could not find it.  This address will work
    # in most places unless Docker changes to use a different address.
    out.warn('Could not find bridge gateway, using default')
    return '172.17.0.1'


def prepare_port_bindings(service):
    bindings = service.requests.get("port-bindings", {}).copy()

    # If the chute is configured to host a web service, check
    # whether there is a host port binding associated with it.
    # If not, we will add blank one so that Docker dynamically
    # assigns a port in the host to forward to the container.
    chute = service.get_chute()
    web_port = chute.web.get("port", None)
    web_service = chute.web.get("service", None)
    if web_port is not None and web_service == service.name:
        # The port could appear in multiple formats, e.g. 80, 80/tcp.
        # If any of them are present, we do not need to add anything.
        keys = [web_port, str(web_port), "{}/tcp".format(web_port)]
        if not any(k in bindings for k in keys):
            bindings["{}/tcp".format(web_port)] = None

    # Check if there is an existing version of the chute installed, so we can
    # potentially inherit old port bindings.
    try:
        container = ChuteContainer(service.get_container_name())
        data = container.inspect()
        old_bindings = data['NetworkSettings']['Ports']
    except ChuteNotFound:
        old_bindings = {}

    # If the current binding is unspecified (None), and there exists previous
    # binding for the same port, we can inherit the previous binding rather
    # than let Docker pick an arbitrary new one.
    #
    # The primary effect is that if a user followed a redirect to the web port
    # of a chute, then updated the chute, the port should remain valid after
    # the update.
    for key, value in six.iteritems(bindings):
        if value is None and key in old_bindings:
            port = old_bindings[key][0]['HostPort']
            bindings[key] = port

    return bindings


def build_host_config(update, service):
    """
    Build the host_config dict for a docker container based on the passed in update.

    :param chute: The chute object containing information about the chute.
    :type chute: obj
    :returns: (dict) The host_config dict which docker needs in order to create the container.
    """
    extra_hosts = {}
    network_mode = service.requests.get('network-mode', 'bridge')
    volumes = update.cache_get('volumes')

    # We are not able to set extra_hosts if the network_mode is set to 'host'.
    # In that case, the chute uses the same /etc/hosts file as the host system.
    if network_mode != 'host':
        extra_hosts[settings.LOCAL_DOMAIN] = getBridgeGateway()

    # If the chute has not configured a host binding for port 80, let Docker
    # assign a dynamic one.  We will use it to redirect HTTP requests to the
    # chute.
    port_bindings = update.cache_get('portBindings:{}'.format(service.name))

    # restart_policy: set to 'no' to prevent Docker from starting containers
    # automatically on system boot.  Paradrop will set up the host environment
    # first, then restart the containers.
    # host_conf = client.create_host_config(
    host_conf = dict(
        cap_add=['NET_ADMIN'],
        cap_drop=[],
        #devices=config.get('devices', []),
        #dns=config.get('dns'),
        dns_search=[],
        extra_hosts=extra_hosts,
        network_mode=network_mode,
        ports=port_bindings,
        privileged=service.requests.get('privileged', False),
        publish_all_ports=False,
        restart_policy={'Name': 'no'},
        volumes=volumes
    )
    return host_conf


def call_retry(cmd, env, delay=3, tries=3):
    # Make sure each component of the command is a string.  Otherwisew we will
    # get errors.
    clean_cmd = [str(v) for v in cmd]

    while tries >= 0:
        tries -= 1

        out.info("Calling: {}\n".format(" ".join(clean_cmd)))
        try:
            proc = subprocess.Popen(clean_cmd, stdout=subprocess.PIPE,
                                    stderr=subprocess.PIPE, env=env)
            for line in proc.stdout:
                out.info("{}: {}\n".format(clean_cmd[0], line.strip()))
            for line in proc.stderr:
                out.warn("{}: {}\n".format(clean_cmd[0], line.strip()))
            return proc.returncode
        except OSError as e:
            out.warn('Command "{}" failed\n'.format(" ".join(clean_cmd)))
            if tries <= 0:
                out.exception(e, True)
                raise e

        time.sleep(delay)


def setup_net_interfaces(update):
    """
    Link interfaces in the host to the internal interfaces in the Docker
    container.

    The commands are based on the pipework script
    (https://github.com/jpetazzo/pipework).

    :param chute: The chute object containing information about the chute.
    :type update: obj
    :returns: None
    """
    interfaces = update.cache_get('networkInterfaces')

    # Construct environment for subprocess calls.
    env = {
        "PATH": os.environ.get("PATH", "/bin")
    }
    if settings.DOCKER_BIN_DIR not in env['PATH']:
        env['PATH'] += ":" + settings.DOCKER_BIN_DIR

    # Keep list of interfaces that assign to the Docker container that we will
    # need to recover when stopping the chute, for example monitor mode
    # interfaces.  They will not automatically return to the default namespace,
    # so we need to keep track of them.
    borrowedInterfaces = []

    for iface in interfaces:
        itype = iface.get('type', 'wifi-ap')

        service = update.new.get_service(iface['service'])

        # We need the container's PID in order to work with Linux namespaces.
        container = ChuteContainer(service.get_container_name())
        pid = container.getPID()

        if itype == 'lan' or itype == 'vlan' or itype == 'wifi-ap':
            IP = iface['ipaddrWithPrefix']
            internalIntf = iface['internalIntf']
            externalIntf = iface['externalIntf']

            # Generate a temporary interface name.  It just needs to be unique.
            # We will rename to the internalIntf name as soon as the interface
            # is inside the chute.
            tmpIntf = "tmp{:x}".format(random.getrandbits(32))

            # TODO copy MTU from original interface?
            cmd = ['ip', 'link', 'add', 'link', externalIntf, 'dev', tmpIntf,
                    'type', 'macvlan', 'mode', 'bridge']
            call_retry(cmd, env, tries=1)

            # Bring the interface up.
            cmd = ['ip', 'link', 'set', tmpIntf, 'up']
            call_retry(cmd, env, tries=1)

            # Give the new interface to the chute.
            cmd = ['ip', 'link', 'set', tmpIntf, 'netns', str(pid)]
            call_retry(cmd, env, tries=1)

            # Rename the interface according to what the chute wants.
            cmd = ['ip', 'link', 'set', tmpIntf, 'name', internalIntf]
            call_in_netns(service, env, cmd)

            # Set the IP address.
            cmd = ['ip', 'addr', 'add', IP, 'dev', internalIntf]
            call_in_netns(service, env, cmd)

            # Bring the interface up again.
            cmd = ['ip', 'link', 'set', internalIntf, 'up']
            call_in_netns(service, env, cmd)

        elif itype == 'wifi-monitor':
            internalIntf = iface['internalIntf']
            externalIntf = iface['externalIntf']
            phyname = iface['phy']

            cmd = ['iw', 'phy', phyname, 'set', 'netns', str(pid)]
            call_retry(cmd, env, tries=1)

            # Rename the interface inside the container.
            cmd = ['ip', 'link', 'set', 'dev', externalIntf, 'up', 'name',
                    internalIntf]
            call_in_netns(service, env, cmd)

            borrowedInterfaces.append({
                'type': 'wifi',
                'pid': pid,
                'internal': internalIntf,
                'external': externalIntf,
                'phy': phyname
            })

        elif itype == '_lan':
            # Not currently supported, this mode would allow chutes to take
            # control of the physical LAN interface directly, rather than the
            # virtual macvlan interface created above.
            internalIntf = iface['internalIntf']
            externalIntf = iface['externalIntf']

            cmd = ['ip', 'link', 'set', 'dev', externalIntf, 'up', 'netns',
                    str(pid), 'name', internalIntf]
            call_retry(cmd, env, tries=1)

            borrowedInterfaces.append({
                'type': 'lan',
                'pid': pid,
                'internal': internalIntf,
                'external': externalIntf
            })

        else:
            raise Exception("Unrecognized interface type: {}".format(itype))

    update.cache_set('borrowedInterfaces', borrowedInterfaces)


def cleanup_net_interfaces(update):
    """
    Cleanup special interfaces when bringing down a container.

    This applies to monitor mode interfaces, which need to be renamed before
    they come back to the host network, e.g. "mon0" inside the container should
    be renamed to the appropriate "wlanX" before the container exits.
    """
    borrowedInterfaces = update.cache_get('borrowedInterfaces')
    if borrowedInterfaces is None:
        return

    # Construct environment for subprocess calls.
    env = {
        "PATH": os.environ.get("PATH", "/bin")
    }
    if settings.DOCKER_BIN_DIR not in env['PATH']:
        env['PATH'] += ":" + settings.DOCKER_BIN_DIR

    for iface in borrowedInterfaces:
        service = update.new.get_service(iface['service'])

        if iface['type'] == 'wifi':
            cmd = ['ip', 'link', 'set', 'dev', iface['internal'], 'down',
                    'name', iface['external']]
            call_in_netns(service, env, cmd, onerror="ignore", pid=iface['pid'])

            cmd = ['iw', 'phy', iface['phy'], 'set', 'netns', '1']
            call_in_netns(service, env, cmd, onerror="ignore", pid=iface['pid'])

            resetWirelessDevice(iface['phy'], iface['external'])

        elif iface['type'] == 'lan':
            cmd = ['ip', 'link', 'set', 'dev', iface['internal'], 'down',
                    'netns', '1', 'name', iface['external']]
            call_in_netns(service, env, cmd, onerror="ignore", pid=iface['pid'])


def call_in_netns(service, env, command, onerror="raise", pid=None):
    """
    Call command within a service's namespace.

    command: should be a list of strings.
    onerror: should be "raise" or "ignore"
    """
    container_name = service.get_container_name()

    if pid is None:
        # We need the chute's PID in order to work with Linux namespaces.
        container = ChuteContainer(container_name)
        pid = container.getPID()

    # Try first with `nsenter`.  This is preferred because it works using
    # commands in the host.  We cannot be sure the `docker exec` version will
    # work with all chute images.
    cmd = ['nsenter', '--target', str(pid), '--net', '--no-fork'] + command
    try:
        code = call_retry(cmd, env, tries=1)
    except:
        code = -1

    # We fall back to `docker exec` which relies on the container image having
    # an `ip` command available to configure interfaces from within.
    if code != 0:
        out.warn("nsenter command failed, resorting to docker exec\n")

        try:
            client = docker.DockerClient(base_url="unix://var/run/docker.sock", version='auto')
            container = client.containers.get(container_name)
            container.exec_run(command, user='root')
        except Exception:
            if onerror == "raise":
                raise


def prepare_environment(update, service):
    """
    Prepare environment variables for a chute container.
    """
    # Make a copy so that we do not alter the original, which only contains
    # user-specified environment variables.
    #
    # Start with the chute environment variables, then apply service-specific
    # variables on top of that.
    env = update.new.get_environment().copy()
    env.update(service.environment)

    env['PARADROP_CHUTE_NAME'] = update.new.name
    env['PARADROP_FEATURES'] = constants.DAEMON_FEATURES
    env['PARADROP_ROUTER_ID'] = nexus.core.info.pdid
    env['PARADROP_DATA_DIR'] = update.cache_get('internalDataDir')
    env['PARADROP_SYSTEM_DIR'] = update.cache_get('internalSystemDir')
    env['PARADROP_API_URL'] = "http://{}/api".format(settings.LOCAL_DOMAIN)
    env['PARADROP_BASE_URL'] = "http://{}/api/v1/chutes/{}".format(
            settings.LOCAL_DOMAIN, update.new.name)
    env['PARADROP_API_TOKEN'] = update.cache_get('apiToken')
    env['PARADROP_WS_API_URL'] = "ws://{}/ws".format(settings.LOCAL_DOMAIN)
    env['PARADROP_CHUTE_VERSION'] = update.new.version

    return env


def _setResourceAllocation(allocation):
    client = docker.DockerClient(base_url="unix://var/run/docker.sock", version='auto')
    for container_name, resources in six.iteritems(allocation):
        out.info("Update chute {} set cpu_shares={}\n".format(
            container_name, resources['cpu_shares']))
        container = client.containers.get(container_name)
        container.update(cpu_shares=resources['cpu_shares'])

        # Using class id 1:1 for prioritized, 1:3 for best effort.
        # Prioritization is implemented in confd/qos.py.  Class-ID is
        # represented in hexadecimal.
        # Reference: https://www.kernel.org/doc/Documentation/cgroup-v1/net_cls.txt
        if resources.get('prioritize_traffic', False):
            classid = "0x10001"
        else:
            classid = "0x10003"

        container = ChuteContainer(container_name)
        try:
            container_id = container.getID()
            fname = "/sys/fs/cgroup/net_cls/docker/{}/net_cls.classid".format(container_id)
            with open(fname, "w") as output:
                output.write(classid)
        except Exception as error:
            out.warn("Error setting traffic class: {}\n".format(error))


def setResourceAllocation(update):
    """
    Adjust compute resources assigned to chute containers.
    """
    allocation = update.cache_get('newResourceAllocation')
    _setResourceAllocation(allocation)


def revertResourceAllocation(update):
    allocation = update.cache_get('oldResourceAllocation')
    _setResourceAllocation(allocation)


def removeAllContainers(update):
    """
    Remove all containers on the system.  This should only be used as part of a
    factory reset mechanism.

    :returns: None
    """
    client = docker.DockerClient(base_url='unix://var/run/docker.sock', version='auto')

    for container in client.containers.list(all=True):
        try:
            container.remove(force=True)
        except Exception as e:
            update.progress(str(e))



import errno
import os
import signal
import six
import subprocess
import time

from paradrop.base.output import out


def kill(pid, kill_signal=4, timeout=8):
    """
    Kill a child process and wait with timeout.

    1. Send a SIGTERM signal to the process.
    2. Wait up to `kill_signal` seconds for the process to exit.
    3. If process is still running, send a SIGKILL signal.
    4. Wait up to `timeout` seconds (cumulative with `kill_signal`) for the
    process to exit.

    Returns True if the process exited before `timeout` seconds elapsed.
    """
    os.kill(pid, signal.SIGTERM)

    start = time.time()
    while (time.time() - start) < kill_signal:
        time.sleep(0.1)

        try:
            # waitpid returns (0, 0) if the process is still running.
            # Otherwise, it returns (pid, status).
            opid, status = os.waitpid(pid, os.WNOHANG)
            if opid == pid:
                return True
        except OSError as error:
            # I think ECHILD means there is no process with the given pid.
            if error.errno == errno.ECHILD:
                return True
            else:
                raise error

    try:
        os.kill(pid, signal.SIGKILL)
    except OSError as error:
        if error.errno == errno.ESRCH:
            # The process exited between the time that we checked it and when
            # we tried to send a SIGKILL.  This is not an error.
            return True
        else:
            raise error

    while (time.time() - start) < timeout:
        time.sleep(0.1)

        try:
            # waitpid returns (0, 0) if the process is still running.
            # Otherwise, it returns (pid, status).
            opid, status = os.waitpid(pid, os.WNOHANG)
            if opid == pid:
                return True
        except OSError as error:
            # I think ECHILD means there is no process with the given pid.
            if error.errno == errno.ECHILD:
                return True
            else:
                raise error

    return False


class CommandList(list):
    def __contains__(self, s):
        """
        Test if the list contains a given string.
        """
        return any(s in cmd for prio, cmd in self)

    def append(self, priority, command):
        super(CommandList, self).append((priority, command))

    def commands(self):
        """
        Iterate over commands in order by priority.

        Commands are first sorted by assigned priority.  Within each priority
        level, the order in which they were added is maintained.
        """
        result = list()
        for i in range(len(self)):
            prio, cmd = self[i]
            result.append((prio, i, cmd))

        result.sort()

        for prio, i, cmd in result:
            yield cmd


class Command(object):
    def __init__(self, command, parent=None, ignoreFailure=False):
        """
        Construct command object.

        command: array of strings specifying command and arguments
                 Passing a single string is also supported if there are
                 no spaces within arguments (only between them).
        parent: parent object (should be ConfigObject subclass)
        """
        self.parent = parent
        self.ignoreFailure = ignoreFailure

        if type(command) == list:
            self.command = [str(v) for v in command]
        elif isinstance(command, six.string_types):
            self.command = command.split()

        # These are set after execute completes.
        self.pid = None
        self.result = None

    def __contains__(self, s):
        """
        Test if command contains given string.

        Example:
        If the cmd.command = ['kill', '1'], then ("kill" in cmd) will return True.
        """
        return (s in str(self))

    def __str__(self):
        return " ".join(self.command)

    def execute(self):
        try:
            proc = subprocess.Popen(self.command, stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE)
            self.pid = proc.pid
            for line in proc.stdout:
                out.verbose("{} {}: {}".format(self.command[0], self.pid, line))
            for line in proc.stderr:
                out.verbose("{} {}: {}".format(self.command[0], self.pid, line))
            self.result = proc.wait()
            if self.result == 0:
                out.verbose('Command "{}" returned {}\n'.format(
                         " ".join(self.command), self.result))
            else:
                out.info('Command "{}" returned {}\n'.format(
                         " ".join(self.command), self.result))
        except Exception as e:
            out.info('Command "{}" raised exception {}\n'.format(
                     " ".join(self.command), e))
            self.result = e

        if self.parent is not None:
            self.parent.executed.append(self)

        return (self.ignoreFailure or self.result == 0)

    def success(self):
        """
        Returns True if the command was successfully executed.
        """
        return (self.ignoreFailure or self.result == 0)


class ErrorCommand(Command):
    """
    Special command object that indicates an error occurred.
    """
    def __init__(self, error, parent=None):
        super(ErrorCommand, self).__init__([], parent)
        self.error = error

    def execute(self):
        out.info("An error occurred: {}".format(self.error))

    def success(self):
        return False


class FunctionCommand(Command):
    """
    Command that runs a Python function.
    """
    def __init__(self, parent, function, *args, **kwargs):
        command = [function.__name__]
        command.extend(args)
        for key, value in six.iteritems(kwargs):
            command.append("{}={}".format(key, value))

        super(FunctionCommand, self).__init__(command, parent)

        self.function = function
        self.args = args
        self.kwargs = kwargs

        self.result = None

    def execute(self):
        self.result = self.function(*self.args, **self.kwargs)


class KillCommand(Command):
    """
    Special command object for killing a process
    """
    def __init__(self, pid, parent=None):
        """
        Create a kill command

        The pid argument can either be a real pid (e.g. kill 12345) or a path
        to a file containing the pid.

        If the pid is coming from a file, it will be resolved at the time that
        execute is called.  Before that time, the command will be stored
        internally as ["kill", "/path/to/file"].  This is not a real command,
        but it is meaningful if you print the command object.
        """
        # This will not be a valid command if pid is a file path.
        command = ["kill", pid]

        super(KillCommand, self).__init__(command, parent)

        # Is it a numeric pid or a path to a pid file?
        try:
            self.pid = int(pid)
            self.fromFile = False
        except ValueError:
            self.pid = pid
            self.fromFile = True

    def getPid(self):
        if self.fromFile:
            try:
                with open(self.pid, "r") as inputFile:
                    return int(inputFile.read().strip())
            except:
                # No pid file --- maybe it was not running?
                out.warn("File not found: {}\n".format(self.pid))
                return None
        else:
            return self.pid

    def execute(self):
        pid = self.getPid()

        if pid is None:
            self.result = 0
            return True

        try:
            retval = kill(pid)
            self.result = 0
            out.info('Command "kill {}" returned {}\n'.format(pid, retval))
        except Exception as e:
            out.info('Command "kill {}" raised exception {}\n'.format(pid, e))
            self.result = e

        return (self.result == 0)



###################################################################
# Copyright 2013-2015 All Rights Reserved
# Authors: The Paradrop Team
###################################################################

from paradrop.base.output import out
from paradrop.core.config import dockerconfig, configservice, dhcp
from . import plangraph


def generatePlans(update):
    """
        This function looks at a diff of the current Chute (in @chuteStor) and the @newChute,
        then adds Plan() calls to make the Chute match the @newChute.

        Returns:
            True: abort the plan generation process
    """
    out.verbose("%r\n" % (update))

    # Generate virt start script, stored in cache (key: 'virtPreamble')
    update.plans.addPlans(plangraph.RUNTIME_GET_VIRT_PREAMBLE, (dockerconfig.getVirtPreamble, ))

    # Make sure volume directories exist.
    update.plans.addPlans(plangraph.CREATE_VOLUME_DIRS,
            (dockerconfig.createVolumeDirs, ),
            (dockerconfig.abortCreateVolumeDirs, ))

    # If the user specifies DHCP then we need to generate the config and store it to disk
    update.plans.addPlans(plangraph.RUNTIME_GET_VIRT_DHCP, (dhcp.getVirtDHCPSettings, ))

    todoPlan = (dhcp.setVirtDHCPSettings, )
    abtPlan = (dhcp.revert_dhcp_settings, )
    update.plans.addPlans(plangraph.RUNTIME_SET_VIRT_DHCP, todoPlan, abtPlan)

    # Reload configuration files
    todoPlan = (configservice.reloadAll, )
    update.plans.addPlans(plangraph.RUNTIME_RELOAD_CONFIG, todoPlan)

    # Reload configuration files if aborting.  This needs to happen at the
    # right place in the update pipeline such that UCI files have been
    # restored to their previous contents.
    todoPlan = (configservice.reload_placeholder, )
    abtPlan = (configservice.reloadAll, )
    update.plans.addPlans(plangraph.RUNTIME_RELOAD_CONFIG_BACKOUT, todoPlan, abtPlan)

    return None



"""
    osconfig module:
        This module is in charge of changing configuration files for
        pdfcd on the host OS. This relates to things like network, dhcp,
        wifi, firewall changes. Pdfcd should be able to make simple abstracted
        calls into this module so that if we need to change what type of OS config
        we need to support only this module would change.
"""

from .uciutils import restoreConfigFile


def revertConfig(update, theType):
    """
    Tell the UCI module to revert changes to the old state of the chute.
    """
    restoreConfigFile(update.new, theType)



from __future__ import print_function
import struct
from datetime import datetime

from twisted.internet.fdesc import setNonBlocking


class SpectrumReader(object):

    # spectral scan packet format constants
    hdrsize = 3
    pktsize = 17 + 56

    # ieee 802.11 constants
    sc_wide = 0.3125  # in MHz

    def __init__(self, path):
        self.fp = file(path)
        if not self.fp:
            raise Exception("Cant open file '%s'" % path)

        setNonBlocking(self.fp.fileno())

    def read_samples(self):
        data = self.fp.read()
        if (data):
            ts = datetime.now()
            return ts, data
        else:
            return None, None

    def flush(self):
        self.fp.read()

    @staticmethod
    def decode(data):
        """
        For information about the decoding of spectral samples see:
        https://wireless.wiki.kernel.org/en/users/drivers/ath9k/spectral_scan
        https://github.com/erikarn/ath_radar_stuff/tree/master/lib
        and your ath9k implementation in e.g.
        /drivers/net/wireless/ath/ath9k/common-spectral.c
        """
        pos = 0
        while pos < len(data) - SpectrumReader.hdrsize + 1:

            (stype, slen) = struct.unpack_from(">BH", data, pos)

            if not (stype == 1 and slen == SpectrumReader.pktsize):
                print("skip malformed packet")
                break  # header malformed, discard data. This event is very unlikely (once in ~3h)
                # On the other hand, if we buffer the sample in a primitive way, we consume too much cpu
                # for only one or two "rescued" samples every 2-3 hours

            # We only support 20 MHz
            if stype == 1:
                if pos >= len(data) - SpectrumReader.hdrsize - SpectrumReader.pktsize + 1:
                    break
                pos += SpectrumReader.hdrsize

                packet = data[pos: pos + SpectrumReader.pktsize]
                pos += SpectrumReader.pktsize

                yield packet

                '''
                (max_exp, freq, rssi, noise, max_mag, max_index, bitmap_weight, tsf) = \
                    struct.unpack_from(">BHbbHBBQ", data, pos)
                pos += 17

                sdata = struct.unpack_from("56B", data, pos)
                pos += 56

                yield (tsf, max_exp, freq, rssi, noise, max_mag, max_index, bitmap_weight, sdata)

                # calculate power in dBm
                sumsq_sample = 0
                samples = []
                for raw_sample in sdata:
                    if raw_sample == 0:
                        sample = 1
                    else:
                        sample = raw_sample << max_exp
                    sumsq_sample += sample*sample
                    samples.append(sample)

                if sumsq_sample == 0:
                    sumsq_sample = 1
                sumsq_sample = 10 * math.log10(sumsq_sample)

                sc_total = 56  # HT20: 56 OFDM subcarriers

                #first_sc = freq - SpectrumReader.sc_wide * (sc_total/2 + 0.5)
                pwr = {}
                for i, sample in enumerate(samples):
                    subcarrier_freq = 0
                    if i < 28:
                        subcarrier_freq = freq - SpectrumReader.sc_wide * (28 - i)
                    else:
                        subcarrier_freq = freq + SpectrumReader.sc_wide * (i - 27)

                    #subcarrier_freq = first_sc + i*SpectrumReader.sc_wide

                    sigval = noise + rssi + 20 * math.log10(sample) - sumsq_sample
                    pwr[subcarrier_freq] = sigval

                yield (tsf, freq, noise, rssi, pwr)
                '''



"""
Resource optimization functions.
"""

#from scipy.optimize import linprog


def allocate(reservations, total=1.0):
    """
    Allocate resources among slices with specified and unspecified reservations.

    Returns a new list of values with the following properties:
    - Every value is >= the corresponding input value.
    - The result sums to `total`.

    Examples:
    allocate([0.25, None, None]) -> [0.5, 0.25, 0.25]
    allocate([0.4, None, None]) -> [0.6, 0.2, 0.2]
    allocate([0.2, 0.2, 0.2]) -> [0.33, 0.33, 0.33]
    allocate([None, None, None]) -> [0.33, 0.33, 0.33]
    allocate([0.5, 0.5, 0.5]) -> ERROR
    """
    n = len(reservations)
    if n <= 0:
        return []

    remaining = total
    for r in reservations:
        if r is not None:
            remaining -= r

    result = list(reservations)

    # Divide the remaining resources among all slices.
    if remaining > 0:
        share = float(remaining) / n
        for i in range(n):
            if result[i] is None:
                result[i] = share
            else:
                result[i] += share
    elif remaining < 0:
        raise Exception("Resource allocation is infeasible (total {} > {})".format(
            sum(result), total))

    return result


#
# The optimize function requires scipy.  It might be useful later.
#

#def optimize(reservations, ub=1):
#    n = len(reservations)
#
#    # In case some reservations are unspecified (None), calculate the remaining
#    # share and divide it evenly among the unspecified buckets.
#    #
#    # Example:
#    # [0.2, None, None] -> [0.2, 0.4, 0.4]
#    floor = allocate(reservations, ub)
#    print(floor)
#
#    # Coefficients for the objective function.  We want to maximize the sum, but
#    # linprog does minimization, hence the negative.
#    c = [-1] * n
#
#    # Constraints (A_ub * x <= b_ub).
#    A_ub = []
#    b_ub = []
#
#    for i in range(n):
#        # Competition constraint (sum of all others <= ub-reservation[i]).
#        # If all others (j != i) use up their allocated resource, there must
#        # be enough left over for reservation[i].
#        row = [0] * n
#        for j in range(n):
#            # Sum of limit[j] for j != i
#            if j != i:
#                row[j] = 1
#
#        # ...is <= ub-reservations[i].
#        A_ub.append(row)
#        b_ub.append(ub-floor[i])
#
#    # Set the upper and lower bounds:
#    # (reservations[i] or fair share) <= x[i] <= ub
#    bounds = [(floor[i], ub) for i in range(n)]
#
#    result = linprog(c, A_ub=A_ub, b_ub=b_ub, bounds=bounds)
#    return result



import os
import shutil

from paradrop.base import settings
from paradrop.core.chute.builder import build_chute
from paradrop.core.container.downloader import downloader
from paradrop.lib.utils import pdosq


def download_chute_files(update):
    """
    Download any files or git repositories required by the chute.
    """
    download_args = getattr(update, "download", None)
    if download_args is not None:
        loader = downloader(**download_args)
        workdir, meta = loader.fetch()
        update.workdir = workdir


def load_chute_configuration(update):
    """
    Load and interpret paradrop.yaml file.
    """
    if not update.has_chute_build():
        return

    config = {}

    workdir = getattr(update, "workdir", None)
    if workdir is not None:
        conf_path = os.path.join(workdir, settings.CHUTE_CONFIG_FILE)
        config = pdosq.read_yaml_file(conf_path, default={})

    # Look for additional build information from the update object
    # and merge with the configuration file.
    if hasattr(update, "build"):
        config.update(update.build)

    # Try to set the owner of the chute.
    if hasattr(update, "user"):
        config['owner'] = update.user

    update.new = build_chute(config)


def delete_chute_files(update):
    """
    Delete files that were used to build the chute.
    """
    workdir = getattr(update, "workdir", None)
    if workdir is not None:
        shutil.rmtree(workdir)
